"""Modèle un peu simpliste de hiérarchie LDAP + info métiers spécifiques."""
from __future__ import annotations

from typing import TYPE_CHECKING

import toolz
from abilian.core.entities import Entity, EntityQuery
from abilian.core.sqlalchemy import JSONDict
from attr import attrib, attrs
from sqlalchemy import Boolean, Column, Enum, ForeignKey, Integer, String, \
    Unicode
from sqlalchemy.orm import backref, foreign, relationship, remote

from labster.domain.services import roles as roles_service
from labster.extensions import db

if TYPE_CHECKING:
    from .profiles import Profile
    from .roles import RoleType

# Constantes utilisées pour typer les OrgUnits
POLE_DE_RECHERCHE = "Pôle de recherche"
UFR = "UFR"
LABORATOIRE = "Laboratoire"
DEPARTEMENT = "Département"
EQUIPE = "Équipe"
BUREAU_DGRTT = "Bureau DGRTT"

TYPE_ENUM = [POLE_DE_RECHERCHE, UFR, LABORATOIRE, DEPARTEMENT, EQUIPE, BUREAU_DGRTT]

__all__ = (
    "OrgUnit",
    "POLE_DE_RECHERCHE",
    "UFR",
    "LABORATOIRE",
    "DEPARTEMENT",
    "EQUIPE",
    "BUREAU_DGRTT",
)


class OrgUnitQuery(EntityQuery):
    def filter_by_type(self, type):
        return self.filter(OrgUnit.type == type)

    def get_by_nom(self, nom):
        return self.filter(OrgUnit.nom == nom).one()

    def get_by_sigle(self, sigle):
        return self.filter(OrgUnit.sigle == sigle).one()

    def get_by_dn(self, dn):
        return self.filter(OrgUnit.dn == dn).one()

    def ufrs_count(self):
        return self.filter_by_type(UFR).count()

    def labos_count(self):
        return self.filter_by_type(LABORATOIRE).count()

    def filter_by_parent(self, parent):
        return self.filter(OrgUnit.parent_id == parent.id)


@attrs(init=False, these={"type": attrib(), "nom": attrib(), "id": attrib()}, hash=True)
class OrgUnit(Entity):
    __tablename__ = "orgunit"
    __indexable__ = False
    query_class = OrgUnitQuery

    type = Column(Enum(*TYPE_ENUM, name="orgunit_type"), nullable=False)

    dn = Column(String, unique=True, index=True)
    nom = Column(Unicode, nullable=False, unique=True)
    sigle = Column(Unicode, default="", nullable=False)

    parent_id = Column(Integer, ForeignKey("orgunit.id"))
    parent = relationship(
        "OrgUnit",
        primaryjoin=remote(Entity.id) == foreign(parent_id),
        backref=backref("children", lazy="joined", cascade="all, delete-orphan"),
    )

    wf_settings = Column(JSONDict(), default=dict)

    permettre_reponse_directe = Column(Boolean)
    permettre_soummission_directe = Column(Boolean)

    def __init__(self, **kw):
        super().__init__(**kw)
        self.wf_settings = {}

    def __unicode__(self):
        return f"<OrgUnit type='{self.type}' nom='{self.nom}' id={self.id}>"

    __str__ = __unicode__

    # def __repr__(self):
    #     return unicode(self).encode('utf8')

    @property
    def sigle_ou_nom(self) -> str:
        return self.sigle or self.nom

    @property
    def depth(self) -> int:
        if self.type == EQUIPE:
            return 4
        if self.type == DEPARTEMENT:
            return 3
        if self.type == LABORATOIRE:
            return 2
        if self.type == UFR:
            return 1
        if self.type == POLE_DE_RECHERCHE:
            return 0
        # Should not happen
        return 0

    @property
    def path(self) -> list[str]:
        t = [""] * 5
        for p in self.parents + [self]:
            t[p.depth] = p.nom
        return t

    @property
    def parents(self) -> list[OrgUnit]:
        p = self
        result = []
        while True:
            p = p.parent
            if not p:
                break
            result.append(p)
        result.reverse()
        return result

    def descendants(self) -> list[OrgUnit]:
        if self.type == EQUIPE:
            return []
        if self.type == DEPARTEMENT:
            return list(self.children)

        children = self.children
        result = list(children)
        for c in children:
            result += c.descendants()
        return result

    def get_contacts_dgrtt(self):
        from .mapping_dgrtt import MappingDgrtt

        return MappingDgrtt.query.get_for_ou(self)

    def get_members_with_role(self, role_type: RoleType | None) -> list[Profile]:
        roles = roles_service.get_roles(role_type=role_type, target=self)
        result = [r.profile for r in roles if r.profile]
        return result

    def get_directeurs(self) -> list[Profile]:
        from .roles import RoleType

        result = self.get_members_with_role(RoleType.DIRECTION)
        assert all(p.has_role("directeur") for p in result)

        # On met le "vrai" directeur en premier
        result = [p for p in result if p.is_directeur] + [
            p for p in result if not p.is_directeur
        ]

        return result

    def get_gestionnaires(self) -> list[Profile]:
        from .roles import RoleType

        result = self.get_members_with_role(RoleType.GDL)
        assert all(p.has_role("gestionnaire") for p in result)
        return result

    def get_administrateurs(self) -> list[Profile]:
        from .roles import RoleType

        result = self.get_members_with_role(RoleType.ALL)
        assert all(p.has_role("all") for p in result)
        return result

    @property
    def direction(self) -> list[Profile]:
        return self.get_directeurs()

    @property
    def gestionnaires(self) -> list[Profile]:
        return self.get_gestionnaires()

    @property
    def administrateurs(self):
        return self.get_administrateurs()

    def set_roles(self, users: list[Profile], role_type: RoleType) -> None:
        roles = roles_service.get_roles(role_type=role_type, target=self)
        for role in roles:
            db.session.delete(role)
        db.session.flush()
        for user in users:
            roles_service.grant_role(user, role_type, self)

    @property
    def directeur(self) -> Profile | None:

        direction = self.get_directeurs()
        direction = [d for d in direction if d.is_directeur]
        return toolz.get(0, direction, None)

    @property
    def adresse(self):
        if self.directeur and self.directeur.adresse:
            return self.directeur.adresse
        return ""

    def validate(self) -> None:
        if self.type == POLE_DE_RECHERCHE:
            assert self.parent is None

        elif self.type == UFR:
            assert self.parent.type == POLE_DE_RECHERCHE

        elif self.type == LABORATOIRE:
            assert self.parent.type in (POLE_DE_RECHERCHE, UFR)

        elif self.type == DEPARTEMENT:
            assert self.parent.type == LABORATOIRE

        elif self.type == EQUIPE:
            assert self.parent.type in (LABORATOIRE, DEPARTEMENT)

        elif self.type == BUREAU_DGRTT:
            assert self.parent is None

        else:
            raise AssertionError()

    def get_labo(self) -> OrgUnit | None:
        if self.type == LABORATOIRE:
            return self
        if not self.parent:
            return None
        if self.parent.type == LABORATOIRE:
            return self.parent
        if self.parent.parent.type == LABORATOIRE:
            return self.parent.parent
        raise AssertionError("Should not happen")

    @property
    def laboratoire(self) -> OrgUnit | None:
        try:
            return self.get_labo()
        except Exception:
            return None

    @property
    def equipe(self) -> OrgUnit | None:
        if self.type == EQUIPE:
            return self
        return None

    @property
    def departement(self) -> OrgUnit | None:
        if self.type == DEPARTEMENT:
            return self
        if self.type == EQUIPE and self.parent.type == DEPARTEMENT:
            return self.parent
        return None

    @property
    def ufr(self) -> OrgUnit | None:
        if self.type == POLE_DE_RECHERCHE:
            return None
        if self.type == UFR:
            return self

        assert self.laboratoire
        labo: OrgUnit = self.laboratoire
        parent = labo.parent
        if parent.type == UFR:
            return parent
        return None

    @property
    def pole(self) -> OrgUnit | None:
        if self.type == POLE_DE_RECHERCHE:
            return self
        if self.type == UFR:
            return self.parent

        assert self.laboratoire
        labo: OrgUnit = self.laboratoire
        parent: OrgUnit = labo.parent
        if parent.type == POLE_DE_RECHERCHE:
            return parent
        return parent.parent

    def get_membres(self) -> list[Profile]:
        if self.type not in [LABORATOIRE, EQUIPE, DEPARTEMENT]:
            return []

        labo = self.get_labo()
        if labo:
            membres_du_labo = {m for m in labo.membres if m.active}
        else:
            membres_du_labo = set()

        if self.type == LABORATOIRE:
            membres = list(membres_du_labo)

        elif self.type == EQUIPE:
            membres = [m for m in membres_du_labo if m.sous_structure == self]

        else:
            # self.type == DEPARTEMENT
            membres_dict = {m for m in membres_du_labo if m.sous_structure == self}
            for equipe in self.children:
                membres_dict.update(equipe.get_membres())
            membres = list(membres_dict)

        def sorter(profile: Profile) -> tuple[str, str]:
            return profile.nom, profile.prenom

        return sorted(membres, key=sorter)

    def wf_must_validate(self, type: str) -> bool:
        assert type in [LABORATOIRE, DEPARTEMENT, EQUIPE]

        if self.type == LABORATOIRE:
            return True

        wf_settings = self.wf_settings
        if type == LABORATOIRE:
            return wf_settings.get("validation_labo", True)
        elif type == DEPARTEMENT:
            return wf_settings.get("validation_dept", True)
        else:
            # type == EQUIPE
            return wf_settings.get("validation_equipe", True)



"""Tests and validation commands."""
from __future__ import annotations

import random

import click
from flask import current_app
from flask.cli import AppGroup, with_appcontext

from labster.domain2.model.demande import DemandeRH
from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure
from labster.domain2.services.calculs_couts import get_ctx_for_demande

from .commands import run_daemons

test = AppGroup("test")


@test.command("testcafe")
@click.option("--headless", is_flag=True)
@with_appcontext
def test_testcafe(headless):
    """Runs the e2e suite using testcafe (old suite)"""
    if headless:
        driver = "chrome:headless"
    else:
        driver = "chrome"

    daemons = [
        ("web", "flask run"),
        ("webpack", "yarn --cwd front run serve"),
        ("test-e2e", f"yarn --cwd front run testcafe {driver} tests/testcafe"),
    ]
    run_daemons(daemons)


@test.command("cypress")
@with_appcontext
def test_cypress():
    """Runs the e2e suite using Cypress (new suite)"""
    daemons = [
        ("web", "flask run"),
        ("webpack", "yarn --cwd front run serve"),
        ("test-e2e", "yarn --cwd front cypress run tests/e2e"),
    ]
    run_daemons(daemons)


@test.command("rpc")
def test_rpc():
    """Runs functional tests on the RPC endpoints."""
    daemons = [
        ("web", "flask run"),
        ("rpc-tests", "pytest rpc_tests"),
    ]
    run_daemons(daemons)


@test.command("validate-db")
@with_appcontext
def test_validatedb():
    def validate_structures():
        structures = Structure.query.all()
        print(f"Validating {len(structures)} structures")
        for ou in structures:
            ou.validate()

    def validate_demandes():
        demandes = DemandeRH.query.all()
        print(f"Validating {len(demandes)} demandes")
        for demande in demandes:
            ctx = get_ctx_for_demande(demande)
            assert ctx

    with current_app.test_request_context():
        validate_structures()
        validate_demandes()


@test.command("crawl")
@with_appcontext
def test_selfcrawl():
    config = current_app.config
    config["TESTING"] = True

    users: list[Profile] = Profile.query.all()
    for user in sorted(random.sample(users, 5000), key=lambda x: x.login):
        if not user.active:
            continue

        print(f"# Crawling with user {user.login}")

        client = current_app.test_client()

        client.get("/backdoor")
        client.get(f"/switch?uid={user.uid}")
        res = client.get("/")
        assert res.status_code == 200



from __future__ import annotations

from flask import current_app
from werkzeug.exceptions import Forbidden, NotFound, abort

from labster.auth import AuthContext
from labster.di import injector
from labster.domain2.model.demande import Demande, DemandeConvention, DemandeRH
from labster.domain2.model.profile import Profile, ProfileRepository
from labster.domain2.model.structure import Structure, StructureRepository
from labster.domain2.model.type_structure import FA, LA
from labster.domain2.services.contacts import ContactService, ContactType
from labster.domain2.services.roles import Role
from labster.domain2.services.workflow.states import EN_EDITION
from labster.ldap.constants import DRI_DN, DRV_DNS
from labster.security import get_current_profile

structure_repo = injector.get(StructureRepository)
profile_repo = injector.get(ProfileRepository)


#
# DR&I
#
def is_membre_dri(user: Profile) -> bool:
    dri = structure_repo.get_by_dn(DRI_DN)
    return user.has_role(Role.MEMBRE, dri)


#
# DRV
#
def is_membre_drv(user: Profile, structure: Structure | None = None) -> bool:
    drv = get_drv_membership(user)
    if not drv:
        return False

    if not structure:
        return True

    fac = drv.parent
    assert fac
    if structure == fac or fac in structure.ancestors:
        return True

    return False


def get_drv_membership(user: Profile) -> Structure | None:
    for drv_dn in DRV_DNS.values():
        drv = structure_repo.get_by_dn(drv_dn)

        # For tests
        if not drv:
            continue

        for role in [Role.MEMBRE, Role.MEMBRE_RATTACHE, Role.MEMBRE_AFFECTE]:
            if user.has_role(role, drv):
                return drv

    return None


#
# Read access
#
def check_read_access(demande: Demande | None):
    """Raises 'Forbidden' if current user doesn't have access to demande."""
    if not demande:
        abort(404)
    if not has_read_access(demande):
        raise Forbidden()


def has_read_access(demande: Demande) -> bool:
    user = get_current_profile()

    if user in (demande.porteur, demande.gestionnaire, demande.contact_labco):
        return True

    if user.has_role(Role.ADMIN_CENTRAL):
        return True

    wf_states = {
        state["new_state"] for state in demande.wf_history if "new_state" in state
    }

    # DRI & DRV
    if wf_states != {EN_EDITION.id}:
        if is_membre_dri(user):
            return True
        if demande.structure and is_membre_drv(user, demande.structure):
            return True

    if is_responsable_ou_gestionnaire(user, demande):
        return True

    if is_responsable_structure_concernee(user, demande):
        return True

    if is_contributeur(user, demande):
        return True

    return False


def is_responsable_ou_gestionnaire(user: Profile, demande: Demande) -> bool:
    if demande.structure:
        structures = [demande.structure] + demande.structure.ancestors
        for structure in structures:
            if user.has_role(Role.RESPONSABLE, structure):
                return True
            if user.has_role(Role.GESTIONNAIRE, structure):
                return True

    return False


def is_gestionnaire(user: Profile, demande: Demande) -> bool:
    if demande.structure:
        structures = [demande.structure] + demande.structure.ancestors
        for structure in structures:
            if user.has_role(Role.GESTIONNAIRE, structure):
                return True

    return False


def is_contributeur(user: Profile, demande: Demande) -> bool:
    for d in demande.data.get("contributeurs", []):
        contributeur_id = d["value"]
        contributeur = profile_repo.get_by_id(contributeur_id)
        if not contributeur:
            continue
        if contributeur == user:
            return True

    return False


def is_responsable_structure_concernee(user: Profile, demande: Demande) -> bool:
    for d in demande.data.get("structures_concernees", []):
        structure_id = d["value"]
        structure = structure_repo.get_by_id(structure_id)
        if not structure:
            continue
        if user.has_role(Role.RESPONSABLE, structure):
            return True

    return False


#
# Write access
#
def check_write_access(demande: Demande | None):
    """Raises 'Forbidden' if current user doesn't have write access to
    demande."""
    if not demande:
        abort(404)
    if not has_write_access(demande):
        raise Forbidden()


def has_write_access(demande: Demande) -> bool:
    if not demande.editable:
        return False

    user = get_current_profile()

    if user in {demande.porteur, demande.gestionnaire}:
        return True

    if is_contributeur(user, demande):
        return True

    return False


def feuille_cout_editable(demande: Demande) -> bool:
    """
    Feuille de coût : celle-ci devient modifiable par le Contact Lab&Co lorsqu'il a la main.
    Chaque enregistrement, par le Porteur, un Contributeur, un Gestionnaire ou un Contact...

    https://trello.com/c/O702eRzQ/
    """
    if not isinstance(demande, DemandeConvention):
        return False

    user = get_current_profile()
    if user in (demande.porteur, demande.gestionnaire, demande.contact_labco):
        return True

    if is_gestionnaire(user, demande):
        return True

    if is_contributeur(user, demande):
        return True

    return False


def check_can_add_pj(demande: Demande | None):
    """Raises 'Forbidden' if current user doesn't have write access to
    demande."""
    if not demande:
        abort(404)
    if not can_add_pj(demande):
        raise Forbidden()


def can_add_pj(demande: Demande) -> bool:
    user = get_current_profile()
    if user in (demande.porteur, demande.gestionnaire, demande.contact_labco):
        return True

    if is_gestionnaire(user, demande):
        return True

    if is_contributeur(user, demande):
        return True

    if user in demande.valideurs():
        return True

    return False


#
# Other (demandes)
#
def can_duplicate(demande):
    user = get_current_profile()
    return user in {demande.gestionnaire, demande.porteur}


def acces_restreint(demande: Demande) -> bool:
    """Retourne True si l'utilisateur courant n'a pas le droit de voir les info
    confidentielles d'un formulaire RH."""

    if not isinstance(demande, DemandeRH):
        return False

    user = get_current_profile()

    if user in {demande.porteur, demande.gestionnaire, demande.contact_labco}:
        return False

    if user.has_role(Role.ADMIN_CENTRAL):
        return False

    # Les responsables "ascendants" ne voient pas le détail
    # sauf pour les structures "concernées".
    # NB: le labo est toujours "concerné"
    if demande.structure:
        if user.has_role(Role.RESPONSABLE, demande.structure):
            return False

    dri = structure_repo.get_by_dn(DRI_DN)
    if user.has_role(Role.RESPONSABLE, dri):
        return False

    structures = demande.get_structure_concernees()
    for structure in structures:
        if user.has_role(Role.RESPONSABLE, structure):
            return False

    # Responsable facultaire et admin facultaire
    if demande.structure:
        structures.add(demande.structure)

    for structure in structures:
        ancestors = structure.ancestors
        for ancestor in ancestors:
            if ancestor.type != FA:
                continue
            if user.has_role(Role.RESPONSABLE, ancestor):
                return False

            # Admin facultaire
            for sous_structure in ancestor.children:
                if sous_structure.sigle.startswith("DRV "):
                    if user.has_role(Role.ADMIN_LOCAL, sous_structure):
                        return False

    contact_service = injector.get(ContactService)
    mapping = contact_service.get_mapping()
    for d in mapping.values():
        if d.get(ContactType.CONTACT_RH) == user:
            return False

    return True


#
# Structures
#
def get_permissions_for_structure(structure: Structure) -> set[str]:
    if current_app.config.get("TESTING"):
        return set()

    user = get_current_profile()
    permissions = set()

    # cas admin central
    if user.has_role(Role.ADMIN_CENTRAL):
        return {"P1", "P2", "P3", "P4", "P5", "P6"}

    # P1: Infos clés - Edition
    # Admin local: Pour S seulement (et ses départements et équipes si S est un Labo)
    # Admin facultaire: De F et de toute structure descendante

    # P2: Hiérarchie des structures – Edition de structures existantes
    # Admin local: Non, à l’exception des départements et équipes si S est un Labo
    # Admin facultaire: De F et de toute structure descendante

    # P3: Hiérarchie des structures – Création / suppression de structures virtuelles (ex : Carnot)
    # Admin local: Non, à l’exception des départements et équipes si S est un Labo
    # Admin facultaire: Non, à l’exception des départements et équipes des laboratoires
    # descendant de F

    # P4: Membre – Edition pour rattachement d’un utilisateur
    # Admin local: Pour S seulement
    # Admin facultaire: Pour F et toute structure descendante

    # P5: Rôles - Edition
    # Admin local: Pour S seulement (et ses départements et équipes si S est un Labo)
    # Admin facultaire: Pour F et toute structure descendante

    # P6: Contacts Lab & Co - Edition
    # Admin local: Non
    # Admin facultaire: Pour F et toute structure descendante

    # Cas admin facultaire
    for ancestor in [structure] + structure.ancestors:
        if user.has_role(Role.ADMIN_LOCAL, ancestor) and ancestor.type == FA:
            permissions.update(["P1", "P2", "P3", "P4", "P5", "P6"])
            return permissions

    # Cas admin local "normal"
    if user.has_role(Role.ADMIN_LOCAL, structure):
        permissions.update(["P3", "P4", "P5"])

    for ancestor in structure.ancestors:
        if user.has_role(Role.ADMIN_LOCAL, ancestor) and ancestor.type == LA:
            permissions.update(["P1", "P2", "P3", "P4", "P5"])

    return permissions


def has_permission(structure: Structure, permission: str) -> bool:
    if current_app.config.get("TESTING"):
        return True

    permissions = get_permissions_for_structure(structure)
    return permission in permissions


def check_permission(structure: Structure | None, permission: str) -> None:
    if not structure:
        raise NotFound()

    if not has_permission(structure, permission):
        raise Forbidden()


def check_structure_editable(structure: Structure | None):
    check_permission(structure, "P1")


def check_can_edit_roles(structure: Structure | None):
    check_permission(structure, "P5")


def check_can_edit_contacts(structure: Structure | None):
    check_permission(structure, "P6")


def is_admin_local(structure: Structure):
    profile = get_current_profile()
    for ancestor in [structure] + structure.ancestors:
        if profile.has_role(Role.ADMIN_LOCAL, ancestor):
            return True
    return False


def can_view_stats():
    auth_context = injector.get(AuthContext)
    user = auth_context.current_profile

    # For tests
    if not user:
        return True

    if user.has_role(Role.RESPONSABLE, "*"):
        return True
    if user.has_role(Role.ADMIN_CENTRAL):
        return True

    return False



"""Modèle un peu simpliste de hiérarchie LDAP + info métiers spécifiques."""
from __future__ import annotations

from abilian.core.entities import Entity
from sqlalchemy import Column, Integer, Unicode


class FaqEntry(Entity):
    __tablename__ = "faq_entry"

    title = Column(Unicode, default="", nullable=False)
    category = Column(Unicode, default="", nullable=False)
    body = Column(Unicode, default="", nullable=False)
    view_count = Column(Integer, default=0)



"""Modèle de case management adapté au projet Labster."""
from __future__ import annotations

from datetime import datetime
from html import escape
from typing import TYPE_CHECKING

from inflection import humanize, underscore

from labster.domain2.services.notifications import send_notification
from labster.types import JSONList

if TYPE_CHECKING:
    from labster.domain2.model.demande import Demande
    from labster.domain2.model.profile import Profile


class WorkflowException(Exception):
    pass


#
# Generic
#
class State:
    next_action = ""
    is_final = False

    @property
    def label(self):
        return humanize(self.__class__.__name__)

    @property
    def label_short(self):
        return self.label

    @property
    def id(self):
        return underscore(self.__class__.__name__).upper()

    def __eq__(self, other):
        return isinstance(other, State) and self.id == other.id

    def __repr__(self):
        return f"<State {self.id}>"

    def enter(self, workflow: Workflow) -> None:
        self.on_leave(workflow)
        workflow.case.wf_state = self.id
        self.on_enter(workflow)
        if self.is_final:
            self.deactivate(workflow)

    def deactivate(self, workflow: Workflow) -> None:
        workflow.case.active = False
        workflow.case.editable = False

    def on_enter(self, workflow: Workflow) -> None:
        """Callback (override in subclasses)."""

    def on_leave(self, workflow: Workflow) -> None:
        """Callback (override in subclasses)."""

    def task_owners(self, workflow: Workflow) -> set[Profile]:
        """Override in subclasses."""
        return set()


class Transition:
    label = ""
    category = "primary"
    from_states: list[State] = []
    to_state: State | None = None
    message = (
        "Transition de l'état '{old_state}' vers l'état "
        "'{new_state}' initiée par l'utilisateur {actor}."
    )
    note = ""

    def precondition(self, workflow: Workflow) -> bool:
        return True

    def apply(self, workflow: Workflow, data: dict) -> None:
        pass

    @property
    def id(self):
        return underscore(self.__class__.__name__).upper()

    def __eq__(self, other):
        return isinstance(other, Transition) and self.id == other.id

    def __repr__(self):
        return f"<Transition {self.id}>"

    def can_execute(self, workflow: Workflow) -> bool:
        if workflow.state not in self.from_states:
            return False
        return self.precondition(workflow)

    def execute(self, workflow: Workflow, data: dict) -> None:
        if not self.can_execute(workflow):
            raise WorkflowException(f"Can't execute transition {self.id}")
        if hasattr(self, "apply"):
            self.apply(workflow, data)
        if self.to_state and not self.to_state == workflow.state:
            self.to_state.enter(workflow)

    def get_users_to_notify(self, workflow: Workflow, old_state: State) -> set[Profile]:
        return set()

    def get_form(self, workflow: Workflow, **kw) -> JSONList:
        return []


class Workflow:
    states: list[State] = []
    transitions: list[Transition] = []
    initial_state: State | None = None
    case: Demande
    actor: Profile

    def __init__(self, case: Demande, actor: Profile):
        self.case = case
        self.actor = actor
        if not case.wf_state:
            self.start()

    def __repr__(self):
        return f"<LabsterWorkflow state={self.case.wf_state}>"

    def set_actor(self, actor):
        self.actor = actor

    def start(self):
        state = self.initial_state
        state.enter(self)

    @property
    def state(self):
        return self.current_state()

    def current_state(self) -> State:
        for state in self.states:
            if self.case.wf_state == state.id:
                return state
        raise WorkflowException(f"Object is in an unknown state: {self.case.wf_state}")

    def current_owners(self) -> set[Profile]:
        return self.current_state().task_owners(self)

    def possible_transitions(self) -> list[Transition]:
        return [t for t in self.transitions if t.can_execute(self)]

    def execute_transition(self, transition: Transition, data: dict | None = None):
        if not data:
            data = {}
        case = self.case
        old_state = self.state
        transition.note = note = data.get("note", "")
        case.no_infolab = data.get("no_infolab", "") or case.no_infolab
        case.no_eotp = data.get("no_eotp", "") or case.no_eotp

        if callable(transition.message):
            message = transition.message(self)
        else:
            message = transition.message
        transition.execute(self, data)
        new_state = self.state
        for k in data:
            if k == "note":
                continue
            self.case.data[k] = data[k]
        self.log(transition, old_state, new_state, message, note)

        self.case.wf_date_derniere_action = datetime.utcnow()
        self.case.wf_retard = 0

    def log(self, transition, old_state, new_state, msg_tpl, note):
        from labster.util import url_for

        actor = self.actor
        ctx = {
            "old_state": old_state,
            "new_state": new_state,
            "actor": f'<a href="{url_for(actor)}">{actor.full_name}</a>',
        }
        msg = msg_tpl.format(**ctx)
        log_entry = {
            "date": datetime.now().strftime("%d %b %Y %H:%M:%S"),
            "actor_id": actor.id,
            "message": msg,
            "note": note,
            "old_state": old_state.id,
            "new_state": new_state.id,
            "transition": transition.id,
        }
        self.case.wf_history.append(log_entry)

        if note:
            notification_msg = msg + f"<blockquote>{escape(note)}</blockquote>"
        else:
            notification_msg = msg

        users_to_notify = transition.get_users_to_notify(self, old_state)
        for user in users_to_notify:
            # FIXME: this shouldn't happen!
            if not user:
                continue
            send_notification(user, notification_msg, self)

    def get_transition_by_id(self, id: str) -> Transition:
        transitions = self.possible_transitions()
        for transition in transitions:
            if transition.id == id:
                return transition
        raise IndexError()

    def get_state_by_id(self, id: str) -> State:
        for state in self.states:
            if state.id == id:
                return state
        raise IndexError()

    def get_value(self, key, default=None):
        return self.case.wf_data.get(key, default)

    def set_value(self, key, value):
        self.case.wf_data[key] = value



from __future__ import annotations

from abilian.core.models import IdMixin
from flask_sqlalchemy import BaseQuery
from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.orm import backref, relationship

from labster.extensions import db

from .profiles import Profile
from .unites import OrgUnit


class MappingDgrttQuery(BaseQuery):
    def get_for_ou(self, ou):
        return self.filter(MappingDgrtt.ou_recherche_id == ou.id).all()


class MappingDgrtt(IdMixin, db.Model):
    __tablename__ = "mapping_dgrtt"
    query_class = MappingDgrttQuery

    ou_recherche_id = Column(Integer, ForeignKey(OrgUnit.id), nullable=False)
    ou_recherche: OrgUnit = relationship(
        OrgUnit,
        foreign_keys=[ou_recherche_id],
        backref=backref("mapping_dgrtt1", cascade="all,delete"),
    )

    bureau_dgrtt = Column(String, nullable=False)

    contact_dgrtt_id = Column(Integer, ForeignKey(Profile.id), nullable=False)
    contact_dgrtt = relationship(
        Profile,
        foreign_keys=[contact_dgrtt_id],
        backref=backref("mapping_dgrtt3", cascade="all,delete"),
    )

    def __repr__(self) -> str:
        return "<MappingDgrtt ou={} bureau={} contact={}>".format(
            self.ou_recherche.sigle_ou_nom, self.bureau_dgrtt, self.contact_dgrtt.uid
        )



from __future__ import annotations

from labster.di import injector
from labster.domain2.model.profile import Profile
from labster.domain2.services.roles import Role, RoleService

role_service = injector.get(RoleService)


def mes_structures(user: Profile):
    roles = role_service.get_roles_for_user(user)
    structures = roles.get(Role.RESPONSABLE, [])
    result = set(structures)
    for s in structures:
        result.update(s.descendants)
    return result



from __future__ import annotations

from labster.domain2.model.demande import _REGISTRY, Demande, \
    DemandeRepository, DemandeRH


def test_single():
    demande = DemandeRH()
    assert isinstance(demande, Demande)


def test_empty_constructors():
    for cls in _REGISTRY.values():
        demande = cls()
        assert isinstance(demande, cls)
        assert isinstance(demande, Demande)


def test_repo(injector, db):
    repo = injector.get(DemandeRepository)

    for cls in _REGISTRY.values():
        demande = cls()
        repo.put(demande)

    repo.clear()



from __future__ import annotations

from datetime import datetime

import pytz
from abilian.core.models import IdMixin
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, Unicode
from sqlalchemy.orm import relationship

from labster.extensions import db

from .demandes import Demande
from .profiles import Profile


class Notification(IdMixin, db.Model):
    __tablename__ = "notification"

    created_at = Column(DateTime, default=datetime.utcnow)

    body = Column(Unicode, nullable=False)

    user_id = Column(Integer, ForeignKey(Profile.id))
    user = relationship(Profile, foreign_keys=[user_id])  # type: Profile

    actor_id = Column(Integer, ForeignKey(Profile.id))
    actor = relationship(Profile, foreign_keys=[actor_id])  # type: Profile

    demande_id = Column(Integer, ForeignKey(Demande.id))
    demande = relationship(Demande, foreign_keys=[demande_id])  # type: Demande

    sent = Column(Boolean)

    @property
    def created_at_tz(self):
        paris = pytz.timezone("Europe/Paris")
        return paris.localize(self.created_at)



from __future__ import annotations

from flask_mail import Mail, Message
from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method
from werkzeug import exceptions

from labster.di import injector
from labster.domain2.model.profile import Profile
from labster.domain.models.faq import FaqEntry
from labster.rpc.queries.faq import check_user_can_edit
from labster.security import get_current_profile

db = injector.get(SQLAlchemy)


@method
def send_message(message):
    mail = injector.get(Mail)
    current_profile = get_current_profile()

    if not message:
        raise exceptions.BadRequest()

    mail.send(make_message(message, current_profile))


@method
def view_entry(id: int):
    entry: FaqEntry = FaqEntry.query.get(id)
    entry.view_count = (entry.view_count or 0) + 1
    db.session.commit()


def make_message(body: str, user: Profile) -> Message:
    subject = "Message reçu sur Lab&co"
    sender = "direction.recherche@upmc.fr"
    body = f"Message de: {user.email}\n\n{body}\n"
    html_body = f"<pre>{body}</pre>"
    recipients = ["direction.recherche@upmc.fr"]
    return Message(
        subject, recipients=recipients, sender=sender, body=body, html=html_body
    )


#
# Admin
#
@method
def update_faq_entry(entry):
    check_user_can_edit()

    entry_id = entry.get("id")
    if entry_id:
        faq_entry = FaqEntry.query.get_or_404(entry_id)
    else:
        faq_entry = FaqEntry()
        db.session.add(faq_entry)

    faq_entry.title = entry["title"]
    faq_entry.body = entry["body"]
    faq_entry.category = entry["category"]

    db.session.commit()


@method
def delete_faq_entry(entry):
    check_user_can_edit()

    entry_id = entry["id"]
    faq_entry = FaqEntry.query.get_or_404(entry_id)
    db.session.delete(faq_entry)
    db.session.commit()



from __future__ import annotations

import uuid
from dataclasses import dataclass

from flask_sqlalchemy import SQLAlchemy
from injector import inject
from sqlalchemy import Column, ForeignKey, String, Table
from sqlalchemy.orm import mapper, relationship

from labster.domain2.model.profile import Profile, ProfileRepository
from labster.domain2.model.structure import Structure, StructureRepository
from labster.domain2.services.contacts import ContactService, ContactType


def new_id() -> str:
    return str(uuid.uuid4())


@dataclass
class Contact:
    id: str
    structure_id: str
    user_id: str
    contact_type_name: str


def make_mapper(metadata):
    table = Table(
        "v3_contacts",
        metadata,
        #
        Column("id", String(36), primary_key=True),
        Column("user_id", String(36), ForeignKey(Profile.id)),
        Column("structure_id", String(36), ForeignKey(Structure.id)),
        Column("contact_type_name", String(64), index=True, nullable=False),
    )

    mapper(
        Contact,
        table,
        properties={
            "structure": relationship(Structure),
            "user": relationship(Profile),
        },
    )


class SqlaContactService(ContactService):
    profile_repo = ProfileRepository
    structure_repo = StructureRepository

    @inject
    def __init__(
        self,
        db: SQLAlchemy,
        profile_repo: ProfileRepository,
        structure_repo: StructureRepository,
    ):
        self.profile_repo = profile_repo
        self.structure_repo = structure_repo
        self.db = db
        self.session = self.db.session
        make_mapper(self.db.metadata)

    def query(self):
        return self.session.query(Contact)

    def is_empty(self):
        return self.query().count() == 0

    def clear(self):
        self.query().delete()
        self.session.flush()

    def get_contact(
        self, structure: Structure, contact_type: ContactType
    ) -> Profile | None:

        assert isinstance(structure, Structure)
        assert isinstance(contact_type, ContactType)

        query = (
            self.query()
            .filter(Contact.structure_id == structure.id)
            .filter(Contact.contact_type_name == contact_type.name)
        )

        contact: Contact | None = query.first()
        if contact:
            return contact.user
        else:
            return None

    def set_contact(
        self, structure: Structure, contact_type: ContactType, user: Profile
    ):
        assert isinstance(structure, Structure)
        assert isinstance(contact_type, ContactType)
        assert isinstance(user, Profile)

        if self.get_contact(structure, contact_type) == user:
            return

        self.delete_contact(structure, contact_type)

        contact = Contact(new_id(), structure.id, user.id, contact_type.name)
        self.db.session.add(contact)
        self.db.session.flush()

    def delete_contact(self, structure: Structure, contact_type: ContactType):
        assert isinstance(structure, Structure)
        assert isinstance(contact_type, ContactType)

        query = (
            self.query()
            .filter(Contact.structure_id == structure.id)
            .filter(Contact.contact_type_name == contact_type.name)
        )

        contact: Contact | None = query.first()
        if contact:
            self.db.session.delete(contact)
            self.db.session.flush()

    def get_mapping_for(self, structure: Structure) -> dict[ContactType, Profile]:
        assert isinstance(structure, Structure)

        query = self.query().filter(Contact.structure_id == structure.id)

        contacts = query.all()
        return {ContactType[c.contact_type_name]: c.user for c in contacts}

    def get_mapping(self) -> dict[Structure, dict[ContactType, Profile]]:
        # TODO: optimize
        query = self.query()
        structures = {c.structure for c in query.all()}

        result = {
            structure: self.get_mapping_for(structure) for structure in structures
        }
        return result

        # result = {}
        # contact_types = list(ContactType)
        #
        # contacts = self.query().all()
        # edges = {
        #     (c.structure_id, ContactType[c.contact_type_name], c.user_id)
        #     for c in contacts
        # }
        #
        # structure_ids = {edge[0] for edge in edges}
        #
        # structures = [
        #     self.structure_repo.get_by_id(StructureId(structure_id))
        #     for structure_id in structure_ids
        # ]
        # structures.sort(key=lambda x: x.name)
        #
        # for structure in structures:
        #     d = {key: None for key in contact_types}
        #     for edge in edges:
        #         if edge[0] == structure.id:
        #             contact_type = edge[1]
        #             user = self.profile_repo.get_by_id(ProfileId(edge[2]))
        #             d[contact_type] = user
        #     result[structure] = d
        #
        # return result



from __future__ import annotations

from typing import Any

import ramda as r
from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method
from marshmallow import Schema, fields
from sqlalchemy import or_

from labster.di import injector
from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure
from labster.domain2.services.roles import Role, RoleService
from labster.rpc.cache import cache
from labster.types import JSONDict
from labster.util import sort_by_name

db = injector.get(SQLAlchemy)
role_service = injector.get(RoleService)


@method
@cache.memoize(tag="users", expire=3600)
def get_users(q="", page=0) -> JSONDict:
    page = int(page)
    query = (
        db.session.query(Profile)
        .filter(Profile.active == True)
        .order_by(Profile.nom, Profile.prenom)
    )
    total = query.count()

    q = q.strip()
    if q:
        query = query.filter(
            or_(Profile.nom.ilike(f"{q}%"), Profile.prenom.ilike(f"{q}%"))
        )

    query = query.offset(100 * page)
    users = query.limit(100).all()

    users_dto = ProfileSchemaMany().dump(users, many=True).data
    return {"users": users_dto, "total": total}


#
# Serialization
#
class ProfileSchemaMany(Schema):
    id = fields.String()
    nom = fields.Function(lambda obj: obj.nom.upper())
    prenom = fields.String()
    structures = fields.Method("get_structures")

    @staticmethod
    def get_structures(user):
        roles = role_service.get_roles_for_user(user)
        return make_structures_dto(roles)


def make_structures_dto(roles_dict: dict[Role, set[Structure]]) -> list[dict]:
    def reverse_dict(d: dict) -> dict[Any, list]:
        result: dict[Any, list] = {}
        for k, l in d.items():
            for v in l:
                result[v] = result.get(v, [])
                result[v].append(k)
        return result

    def make_structure_dto(structure):
        role_set = {role.value for role in structures_to_roles[structure]}
        role_set -= {"Membre", "Membre affilié"}
        if "Membre affecté" in role_set or "Membre rattaché" in role_set:
            role_set.add("Membre")
        role_set -= {"Membre affecté", "Membre rattaché"}
        roles = sorted(role_set)

        structure_dto = {
            "name": structure.sigle_ou_nom,
            "id": structure.id,
            "roles": ", ".join(roles),
            "depth": structure.depth,
        }
        return structure_dto

    structures_to_roles = reverse_dict(roles_dict)
    structures = sort_by_name(structures_to_roles.keys())
    structures_dto = r.pipe(r.map(make_structure_dto), r.filter(lambda s: s["roles"]))(
        structures
    )
    structures_dto.sort(key=lambda x: x["depth"])
    return structures_dto



from __future__ import annotations

from functools import singledispatch
from typing import Any

import structlog
from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method
from marshmallow.fields import Method
from marshmallow_sqlalchemy import TableSchema
from more_itertools import first

from labster.di import injector
from labster.domain2.model.demande import Demande, DemandeRH
from labster.domain2.model.profile import Profile
from labster.domain2.services.roles import Role, RoleService
from labster.infrastructure.repositories.sqla.mappers import Mapper
from labster.newforms import get_form_class_by_name, get_form_class_for
from labster.rbac import acces_restreint, can_duplicate, check_read_access, \
    feuille_cout_editable, has_write_access
from labster.rpc.registry import context_for
from labster.rpc.util import owner_sorter
from labster.security import get_current_profile
from labster.types import JSON, JSONDict, JSONList

logger = structlog.get_logger()
mapper = injector.get(Mapper)
db = injector.get(SQLAlchemy)
role_service = injector.get(RoleService)


@context_for("demande.new")
@method
def get_new(type: str = "calculette_rh") -> JSONDict:
    """Retourne les données nécessaires pour un formulaire vierge."""
    calculette = False
    if type == "calculette_rh":
        calculette = True
        type = "rh"
    form_class = get_form_class_by_name(type)
    form = form_class(mode="create")
    form.update_choices()

    if calculette:
        hide_fields(form)

    model = form.empty_model()
    post_process_form_and_model(form, model)
    return {"form": form.to_dict(), "model": model}


@method
def get_demande(id: str) -> JSON:
    demande = db.session.query(Demande).get(id)
    check_read_access(demande)

    form = get_form_class_for(demande)()
    form.errors = False  # FIXME later
    post_process_form_and_model(form)

    demande_dto = DemandeSchema().dump(demande).data
    demande_dto = add_more_fields(demande, demande_dto)

    if demande_dto["is_editable"]:
        form.update_choices()

    return {"demande": demande_dto, "form": form.to_dict()}


#
# Util
#
def post_process_form_and_model(form, model=None):
    user = get_current_profile()
    roles = role_service.get_roles_for_user(user)
    structures_d_affectation = roles[Role.MEMBRE_AFFECTE]

    structures_comme_gestionnaire = roles[Role.GESTIONNAIRE]
    structures_comme_porteur = roles[Role.PORTEUR]

    structures = []
    if structures_comme_gestionnaire:
        structures += structures_comme_gestionnaire
    if structures_comme_porteur:
        structures += structures_comme_porteur

    field = form.get_field("laboratoire")
    field.choices = [{"value": s.id, "label": s.nom} for s in structures]

    porteurs = {user}
    for structure in structures_comme_gestionnaire:
        membres = role_service.get_users_with_given_role(Role.PORTEUR, structure)
        porteurs.update(membres)

    field = form.get_field("porteur")
    choices = [
        {"value": user.id, "label": f"{user.nom}, {user.prenom}"}
        for user in sorted(porteurs, key=lambda x: (x.nom, x.prenom))
    ]
    field.choices = choices

    if not model:
        return

    porteur = user
    model["porteur"] = {"value": porteur.id, "label": f"{user.nom}, {user.prenom}"}

    structure = first(structures_d_affectation)
    model["laboratoire"] = {"value": structure.id, "label": structure.nom}


#
# Helper for calculette
#
def hide_fields(form):
    fieldsets = form.fieldsets
    get_fieldset(fieldsets, "laboratoire").hidden = True
    get_fieldset(fieldsets, "responsable_scientifique").hidden = True
    get_fieldset(fieldsets, "contrat").hidden = True
    get_fieldset(fieldsets, "structures_concernees").hidden = True

    # Afficher "experience professionnelle", pour montrer le salaire indicatif.
    # fieldsets 4, 5: experience professionnelle.
    get_fieldset(fieldsets, "candidat").hidden = False  # but show just two fieldsets
    get_fieldset(fieldsets, "candidat").fields[0].hidden = True
    get_fieldset(fieldsets, "candidat").fields[1].hidden = True
    get_fieldset(fieldsets, "candidat").fields[2].hidden = True
    get_fieldset(fieldsets, "candidat").fields[3].hidden = True

    get_fieldset(fieldsets, "autre_modification").hidden = True
    get_fieldset(fieldsets, "publicite").hidden = True
    get_fieldset(fieldsets, "commentaire").hidden = True

    get_fieldset(fieldsets, "poste").fields[0].hidden = True
    get_fieldset(fieldsets, "poste").fields[1].hidden = True
    # equivalent corps/grade
    get_fieldset(fieldsets, "poste").fields[2].hidden = False
    get_fieldset(fieldsets, "poste").fields[3].hidden = True


def get_fieldset(fieldsets, name):
    """Return a fieldset."""
    for it in fieldsets:
        if it.name == name:
            return it

    raise RuntimeError(
        f"Could not find fieldset {name}. This is not supposed to happen."
    )


#
# Serialization
#
class DemandeSchema(TableSchema):
    class Meta:
        # model = Demande
        table = mapper.metadata.tables["v3_demandes"]

        include = {
            # Permissions
            "acces_restreint": Method("get_acces_restreint"),
            "is_editable": Method("get_is_editable"),
            "is_duplicable": Method("get_is_duplicable"),
            "feuille_cout_editable": Method("get_feuille_cout_editable"),
            # Key users
            "contact_labco": Method("get_contact_labco"),
            "porteur": Method("get_porteur"),
            "gestionnaire": Method("get_gestionnaire"),
            "contributeurs": Method("get_contributeurs"),
            # Structure
            "structure": Method("get_structure"),
            # Form
            "form_data": Method("get_form_data"),
            # For tabs
            "pieces_jointes": Method("get_pieces_jointes"),
            "workflow": Method("get_workflow"),
            "workflow_history": Method("get_history"),
            "past_versions": Method("get_past_versions"),
            "is_valid": Method("get_is_valid"),
            "errors": Method("get_errors"),
            "extra_errors": Method("get_extra_errors"),
        }

    # Permissions
    def get_acces_restreint(self, demande: Demande) -> bool:
        return acces_restreint(demande)

    def get_is_editable(self, demande: Demande) -> bool:
        return has_write_access(demande)

    def get_is_duplicable(self, demande: Demande) -> bool:
        return can_duplicate(demande)

    def get_feuille_cout_editable(self, demande: Demande) -> bool:
        return feuille_cout_editable(demande)

    # Serialize key users
    def get_contact_labco(self, demande: Demande):
        return self.get_user_field(demande, "contact_labco")

    def get_porteur(self, demande: Demande):
        return self.get_user_field(demande, "porteur")

    def get_gestionnaire(self, demande: Demande):
        return self.get_user_field(demande, "gestionnaire")

    def get_user_field(self, demande: Demande, key):
        user = getattr(demande, key)
        return {"full_name": user.full_name, "id": user.id}

    def get_contributeurs(self, demande: Demande):
        contributeur_ids = {x["value"] for x in demande.data.get("contributeurs", [])}
        contributeurs = set()
        for id in contributeur_ids:
            contributeur = db.session.query(Profile).get(id)
            if contributeur and contributeur.active:
                contributeurs.add(contributeur)
        return [{"full_name": p.full_name, "id": p.id} for p in contributeurs]

    def get_structure(self, demande: Demande):
        structure = demande.structure
        if structure:
            return {"id": structure.id, "label": structure.nom}
        else:
            return {}

    # Form
    def get_form_data(self, demande: Demande):
        return dict(demande.data)

    def get_errors(self, demande: Demande):
        return demande.errors

    def get_extra_errors(self, demande: Demande):
        return demande.get_extra_errors()

    # Tabs
    def get_workflow(self, demande: Demande) -> JSON:
        current_user = get_current_profile()
        workflow = demande.get_workflow(current_user)

        owners = list(workflow.current_owners())

        owners.sort(key=owner_sorter)
        owners_dto = [
            {"full_name": owner.full_name, "id": owner.id} for owner in owners
        ]

        transitions_dto = [
            {
                "id": t.id,
                "label": t.label,
                "category": t.category,
                "form": t.get_form(workflow),
            }
            for t in workflow.possible_transitions()
        ]

        result = {
            "state": {
                "label": workflow.state.label,
                "next_action": workflow.state.next_action,
            },
            "owners": owners_dto,
            "history": [],  # TODO
            "transitions": transitions_dto,
        }
        return result

    # Pour l'onglet "Pièces à joindre"
    def get_pieces_jointes(self, obj: Demande) -> JSONList:
        result: list[dict[str, Any]] = []
        for name, v in obj.attachments.items():
            creator_login = v["creator"]
            creator = (
                db.session.query(Profile).filter(Profile.login == creator_login).first()
            )
            if creator:
                creator_dto = {"id": creator.id, "name": creator.full_name}
            else:
                creator_dto = {"id": None, "name": "Utilisateur inconnu"}

            result.append(
                {
                    "name": name,
                    "id": v["id"],
                    "creator": creator_dto,
                    "date": v.get("date", None),
                }
            )
        return result

    # Pour l'onglet "Documents générés"
    def get_is_valid(self, obj: Demande) -> JSON:
        return obj.is_valid()

    # Pour l'onglet "historique"
    def get_past_versions(self, obj: Demande) -> JSON:
        return obj.past_versions or []

    def get_history(self, obj: Demande) -> JSON:
        return obj.wf_history

    # Workflow
    def get_possible_transitions(self, obj: Demande) -> JSON:
        user = get_current_profile()
        return obj.get_workflow(user).possible_transitions()


@singledispatch
def add_more_fields(demande: Demande, dto: JSONDict):
    return dto


@add_more_fields.register
def add_more_fields_rh(demande: DemandeRH, dto: JSONDict):
    dto["salaire_brut_mensuel"] = demande.data.get("salaire_brut_mensuel", None)
    dto["date_debut"] = demande.data.get("date_debut", None)
    dto["duree_mois"] = demande.duree_mois
    if not dto["acces_restreint"]:
        try:
            dto["cout_total_charge"] = float(demande.cout_total_charge)
        except ValueError:
            dto["cout_total_charge"] = 0
    return dto



"""Formulaire pour une DemandeAutre."""
from __future__ import annotations

from labster.newforms.base import FieldSet, Form, Select2Field, StringField, \
    TextAreaField

#
# Fieldsets
#
from .common import laboratoire

TYPE_CHOICES = [
    "Budget",
    "Plateformes: calcul des coûts complets, réalisation du tarif, passage en conseil…",
    "Convention d'hébergement de start-ups, d'associations…",
    "GDR, GIS, structures en réseaux…",
    "Autre",
]

porteur = FieldSet(  #
    "porteur",
    "Porteur Lab&Co",
    [Select2Field("porteur", "Porteur", choices=[], required=True)],
)

demande = FieldSet(  #
    "demande",
    "Demande",
    [
        StringField("titre", "Titre", required=True),
        Select2Field("type", "Type", choices=TYPE_CHOICES, required=True),
        TextAreaField("commentaire", "Commentaire"),
    ],
)


#
# The Form
#
class DemandeAutreForm(Form):
    name = "autre"
    template = "newforms/newform.html"
    label = "Autre Demande"

    fieldsets = [laboratoire, porteur, demande]



from __future__ import annotations

from typing import Any

from flask_sqlalchemy import SQLAlchemy
from injector import inject
from sqlalchemy.orm import Session

from labster.domain2.model.structure import Structure, StructureRepository
from labster.infrastructure.repositories.sqla.mappers import Mapper


class SqlaStructureRepository(StructureRepository):
    session: Session

    @inject
    def __init__(self, db: SQLAlchemy, mapper: Mapper):
        self.db = db
        self.session = db.session

    def query(self):
        return self.session.query(Structure).filter(Structure.active == True)

    def get_all(self) -> set[Structure]:
        return set(self.query().all())

    def put(self, structure: Structure):
        self.session.add(structure)
        self.session.flush()

    def delete(self, structure: Structure):
        self.session.delete(structure)
        self.session.flush()

    def is_empty(self):
        return self.query().count() == 0

    def clear(self):
        self.query().delete()

    def get_by(self, key: str, value: Any) -> Structure | None:
        return self.query().filter_by(**{key: value}).first()

    # Not needed I think
    # def get_by_id(self, id: StructureId) -> Optional[Structure]:
    #     return self.query().get(id)
    #
    # def get_by_dn(self, dn: str) -> Optional[Structure]:
    #     return self.query().filter(Structure.dn == dn).first()



from __future__ import annotations

from .fields import HTML, Boolean2Field, BooleanField, DateField, EmailField, \
    FieldSet, IntegerField, ListeAutresDeclarations, ListeContrats, \
    ListeDivulgationsFutures, ListeDivulgationsPassees, \
    ListeLicencesExistantes, ListeMateriels, ListePartenaires, \
    ListePartenairesContactes, Select2Field, StringField, TextAreaField, \
    TriStateField
from .form import Form



# from __future__ import annotations
#
# from collections import defaultdict
# from typing import Any, Dict, Set, Tuple
#
# from injector import inject
#
# from labster.domain2.model.profile import Profile, ProfileId, ProfileRepository
# from labster.domain2.model.structure import Structure, StructureId, \
#     StructureRepository
# from labster.domain2.services.roles import Role, RoleService
#
#
# class InmemoryRoleService(RoleService):
#     grants: Set[Tuple[str, str, str]]
#     profile_repository: ProfileRepository
#     structure_repository: StructureRepository
#
#     @inject
#     def __init__(
#         self,
#         profile_repository: ProfileRepository,
#         structure_repository: StructureRepository,
#     ):
#         self.grants = set()
#         self.profile_repository = profile_repository
#         self.structure_repository = structure_repository
#
#     def get_state(self):
#         return [self.grants]
#
#     def set_state(self, state):
#         self.grants = state[0]
#
#     def is_empty(self):
#         return len(self.grants) == 0
#
#     def clear(self):
#         self.grants.clear()
#
#     def grant_role(self, user: Profile, role: Role, context: Any = None):
#         assert isinstance(user, Profile)
#         assert isinstance(role, Role)
#
#         if context:
#             entry = (user.id, role.name, context.id)
#         else:
#             entry = (user.id, role.name, None)
#         self.grants.add(entry)
#
#     def ungrant_role(self, user: Profile, role: Role, context: Any = None):
#         assert isinstance(user, Profile)
#         assert isinstance(role, Role)
#
#         if context:
#             entry = (user.id, role.name, context.id)
#         else:
#             entry = (user.id, role.name, None)
#         assert entry in self.grants
#         self.grants.remove(entry)
#
#     def has_role(self, user: Profile, role: Role, context: Any = None):
#         assert isinstance(user, Profile)
#         assert isinstance(role, Role)
#
#         if context:
#             return (user.id, role.name, context.id) in self.grants
#         else:
#             return (user.id, role.name, None) in self.grants
#
#     def get_users_with_role(self, role: Role, context: Any = None):
#         assert isinstance(role, Role)
#
#         if context:
#             must_match = (role.name, context.id)
#         else:
#             must_match = (role.name, None)
#         users = set()
#         for entry in self.grants:
#             user_id, role_name, context1_id = entry
#             if (role_name, context1_id) == must_match:
#                 user = self.profile_repository.get_by_id(ProfileId(user_id))
#                 users.add(user)
#
#         assert all(self.has_role(u, role, context) for u in users)
#         return users
#
#     def get_users_with_role_on(self, context: Structure) -> Dict[Role, Set[Profile]]:
#         result: Dict[Role, Set] = defaultdict(set)
#         for entry in self.grants:
#             user_id, role_name, context1_id = entry
#             if context1_id == context.id:
#                 user = self.profile_repository.get_by_id(ProfileId(user_id))
#                 role = getattr(Role, role_name)
#                 result[role].add(user)
#         return result
#
#     def get_users_with_given_role(self, role: Role, context: Structure):
#         assert isinstance(role, Role)
#
#         structures_id_to_match = {context.id}
#
#         users = set()
#         for entry in self.grants:
#             user_id, role_name, context1_id = entry
#             if role_name == role.name and context1_id in structures_id_to_match:
#                 user = self.profile_repository.get_by_id(ProfileId(user_id))
#                 users.add(user)
#
#         return users
#
#     def get_roles_for_user(self, user: Profile) -> Dict[Role, Set[Structure]]:
#         result: Dict[Role, Set[Structure]] = defaultdict(set)
#         for entry in self.grants:
#             user_id, role_name, context_id = entry
#             if user_id == user.id:
#                 context = self.structure_repository.get_by_id(StructureId(context_id))
#                 if not context:
#                     continue
#                 role = getattr(Role, role_name)
#                 result[role].add(context)
#
#         return result
from __future__ import annotations



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['labster',
 'labster.bi',
 'labster.blueprints',
 'labster.blueprints.auth',
 'labster.blueprints.bi',
 'labster.blueprints.logging',
 'labster.blueprints.notifications',
 'labster.blueprints.v3',
 'labster.blueprints.v3.user',
 'labster.bus',
 'labster.cli',
 'labster.domain',
 'labster.domain.models',
 'labster.domain.services',
 'labster.domain2',
 'labster.domain2.model',
 'labster.domain2.model.tests',
 'labster.domain2.services',
 'labster.domain2.services.tests',
 'labster.domain2.services.workflow',
 'labster.extensions',
 'labster.forms',
 'labster.infrastructure',
 'labster.infrastructure.repositories',
 'labster.infrastructure.repositories.inmemory',
 'labster.infrastructure.repositories.sqla',
 'labster.infrastructure.services',
 'labster.infrastructure.services.inmemory',
 'labster.infrastructure.services.sqla',
 'labster.ldap',
 'labster.lib',
 'labster.newforms',
 'labster.newforms.base',
 'labster.rpc',
 'labster.rpc.bi',
 'labster.rpc.commands',
 'labster.rpc.queries',
 'labster.rpc.tests',
 'labster.test']

package_data = \
{'': ['*'],
 'labster': ['templates/*', 'templates/emails/*'],
 'labster.blueprints.auth': ['templates/auth/*'],
 'labster.blueprints.bi': ['templates/bi/*'],
 'labster.blueprints.logging': ['templates/logs/*'],
 'labster.blueprints.notifications': ['templates/notifications/debug/*',
                                      'templates/notifications/email/*']}

install_requires = \
['Flask-WeasyPrint',
 'abilian-core>=0.11.12',
 'apscheduler>=3.6.3,<4.0.0',
 'attrs',
 'cairocffi>=1.0.2,<2.0.0',
 'devtools',
 'diskcache>=4.1,<5.0',
 'dotted',
 'email_validator',
 'flask-assets<2',
 'flask-babel<2',
 'flask-injector',
 'flask-profiler>=1.8,<2.0',
 'flask-redis',
 'flask-smorest',
 'glom',
 'gunicorn<20',
 'html2text',
 'inflection',
 'invoke>=1.3,<2.0',
 'iso8601',
 'jsonrpcserver>=4.0,<5.0',
 'ldif>=4,<5',
 'loguru>=0.6.0,<0.7.0',
 'markupsafe>=2,<2.1',
 'marshmallow-annotations',
 'marshmallow-sqlalchemy',
 'marshmallow<3',
 'more-itertools',
 'nox',
 'pampy>=0.3.0,<0.4.0',
 'pandas>=1.2,<2.0',
 'psycopg2',
 'python-dotenv>=0.10.1,<0.11.0',
 'ramda>=0.5.2,<0.6.0',
 'requests[security]>=2.13,<3.0',
 'structlog',
 'toolz',
 'types-python-dateutil>=2.8.10,<3.0.0',
 'types-pytz>=2021.3.6,<2022.0.0',
 'types-requests>=2.27.14,<3.0.0',
 'types-setuptools>=57.4.11,<58.0.0',
 'werkzeug<1',
 'wtforms<3']

entry_points = \
{'console_scripts': ['labster = labster.cli:cli']}

setup_kwargs = {
    'name': 'labandco',
    'version': '2.2.32',
    'description': 'solution métier pour fluidifier les relations entre chercheurs et administration dans les universités et les EPST',
    'long_description': '# Lab&Co - solution métier pour fluidifier les relations entre chercheurs et administration\n\n## A propos de Lab&Co\n\nLab&Co est un progiciel permettant aux établissements de recherche ou d\'enseignement et de recherche(universités, EPST, etc.) de fluidifier les relations entre les chercheurs (porteurs de projets) et l\'administration.\n\nIl couvre notamment:\n\n- Le recrutement de personnel sur contrat de projets (doctorants, bourses, CDD...)\n- Les contrats de recherche (recherche partenariale, ANR, projets européens...)\n- La valorisation de la recherche (logiciels, brevets, bases de donnée...)\n\n## Illustration\n\n<img src="https://raw.githubusercontent.com/abilian/labandco/main/doc/illustration.png">\n\n### Historique\n\nLab&Co est développé par [Abilian](http://www.abilian.com/) en partenariat avec [Will Strategy](https://willstrategy.com/) depuis 2014 pour l\'Université Pierre et Marie Curie (UPMC), devenue depuis Sorbonne Université.\n\nIl est utilisé par plusieurs centaines de chercheurs et la soixantaine de personnes travaillant à la DR&I (Direction de la Recherche et de l\'Innovation).\n\nLa version du code que vous consultez actuellement est la version "3.0" de l\'application, qui été déployée fin 2019 à Sorbonne Université.\n\n### Votre projet\n\nLab&Co peut être déployé dans votre établissement, après connexion à son annuaire LDAP et adaptation des modèles et des workflows à vos processus métiers.\n\nContactez-nous à: [contact@abilian.com](mailto:contact@abilian.com) pour toute demande d\'information.\n\n\n## Développement\n\n### Installer en 1 ligne\n\nSi votre machine de développement est "raisonnablement" configurée, vous pouvez installer Lab&Co localement en une ligne avec la commande:\n\n    ./install.py\n\nDe manière alternative, si vous avez [Poetry](https://poetry.eustace.io/) installé, vous pouvez tapper:\n\n    poetry install\n\n\n### Installer un environnement de développement\n\n1) Créer et activer un virtualenv (Python 3.9 pour l\'instant)\n\n2) Installer Poetry <https://github.com/sdispater/poetry>\n\n3) Dans `/etc/hosts`, faire pointer `labster.local` sur `localhost`.\n\n4) Tapper:\n\n        poetry install\n        yarn --cwd front\n\n\n### Comment développper\n\nEn mode développement, il suffit de lancer `flask devserver`.\n\nPuis lancer le navigateur sur: http://labster.local:5000/\n\nPour lancer les tests: `pytest` ou mieux `tox`.\n\nIl y d\'autres commandes utiles, tapper juste `flask` pour avoir la\nliste.\n\n\n### BDD de prod\n\nCréer une base Postgresql `labster`:\n\n    createdb labster\n\npuis:\n\n    flask create-db\n\n\n### Migrations\n\n    flask db migrate\n    flask db upgrade\n\nDoc: http://alembic.zzzcomputing.com/en/latest/tutorial.html\n\n\n### Tests\n\n1) Tests unitaires:\n\n        pytest\n        yarn --cwd front run test:unit\n\n2) Tests e2e:\n\n        flask cypress\n        flask testcafe\n\n3) Pour tester manuellement en tant qu\'un autre utilisateur, utiliser l\'url\n`backdoor/uid=xxx` (see blueprints/auth/cas.py).\n\n\n#### Tester l\'envoi de mails\n\nLancer\n\n    python -m smtpd -n -c DebuggingServer localhost:1025\n\net dans `DevConfig`\n\n    # MAIL_SUPPRESS_SEND = True\n    MAIL_SUPPRESS_SEND = False\n    MAIL_SERVER = \'localhost\'\n    MAIL_PORT = 1025\n\nAinsi les mais envoyés seront affichés sur stdout.\n\nAutre possibilité, mettre la variable que l\'on veux changer (par rapport au défaut)\ndans un fichier `.env` (similaire au fichier `.flaskenv` qui existe déjà).\n\nLe fichier `.env` a priorité sur `.flaskenv`.\n\nPour les tests unitaires, voir la doc de [flask-mail](https://pythonhosted.org/Flask-Mail/).\n\n\n## Architecture\n\nL\'application est constitué d\'un backend en Python, utilisant notamment le framework Web Flask, l\'ORM SQLAlchemy, et d\'un front-end SPA en JavaScript utilisant le framework réactif Vuejs.\n\nLa communication entre front-end et back-end se fait soit en utilisant le paradigme REST (sans HATEOAS), soit le protocole JSON-RPC 2.0.\n\nLe backend utilise également le framework Abilian-Core développé par Abilian. Il utilise également l\'injection de dépendance grâce à la bibliothèque Injector (+ Flask-Injector).\n\nL\'authentification se fait en utilisant le protocole CAS implémenté par l\'Université.\n\n### Modèle de données (simplifié)\n\n<img src="https://raw.githubusercontent.com/abilian/labandco/main/doc/model.png">\n\n\n\n## FAQ\n\n### L\'application s\'appelle-t-elle "Labster" ou "Lab&Co" ?\n\nLe projet s\'appelait initialement "Labster", d\'où le nom encore donné au package principal et à certaines variables ou constantes. Il s\'appelle à présent "Lab&Co".\n',
    'author': 'Abilian SAS',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import structlog
from abilian.core.extensions import csrf, db, login_manager
from flask_redis import FlaskRedis
from flask_smorest import Api
from werkzeug.exceptions import abort

from .mail import Mail
from .whoosh import Whoosh

logger = structlog.get_logger()

redis = FlaskRedis()
mail = Mail()
api = Api()
whoosh = Whoosh()

db = db

# Ensure that these import are not removed by PyCharm
assert db
assert login_manager


@csrf.error_handler
def csrf_error_response(reason):
    # let sentry be aware of csrf failures. They might show app is broken
    # somewhere
    logger.error("Csrf error report, reason: %s", reason, extra={"stack": True})
    return abort(400, reason)



""""""
from __future__ import annotations

import uuid

import click
import structlog
from abilian.cli import script
from flask.cli import AppGroup, FlaskGroup
from werkzeug.utils import import_string

from labster.app import create_app

# from . import commands, restheart, testing

MODULE_NAMES = ["commands", "restheart", "testing", "indexing"]


#
# Register commands and shell-level CLI
#
@click.group(cls=FlaskGroup, create_app=create_app)
def cli():
    """Management script for the labster application."""


def register_module(namespace, module_name):
    fqn = "labster.cli." + module_name
    try:
        module = import_string(fqn)
        namespace.update(vars(module))
    except ImportError as e:
        print(f"Failed to import module: {fqn}")
        print(e)


def register_commands(app):
    request_id = str(uuid.uuid4())
    logger = structlog.get_logger(request_id=request_id, path="[CLI]")

    logger.debug("registering commands")

    namespace = {}
    for module_name in MODULE_NAMES:
        register_module(namespace, module_name)

    app.cli.add_command(cli)
    app.cli.add_command(script)

    for _name, obj in sorted(namespace.items()):
        if isinstance(obj, (click.core.Command, AppGroup)):
            app.cli.add_command(obj)



from __future__ import annotations

from datetime import datetime
from smtplib import SMTPException
from typing import TYPE_CHECKING

import structlog
from flask import render_template
from flask_mail import Message

from labster.extensions import db, mail

if TYPE_CHECKING:
    from labster.domain2.model.notification import Notification
    from labster.domain2.model.profile import FLUX_TENDU, Profile
    from labster.lib.workflow import Workflow


logger = structlog.get_logger()


def send_notification(user: Profile, body: str, workflow: Workflow) -> Notification:
    """
    :param user: user to notify
    :param body: message body (HTML)
    :param workflow:

    :return: the notification object
    """

    from labster.domain2.model.notification import Notification
    from labster.domain2.model.profile import FLUX_TENDU

    notification = Notification(
        user=user, body=body, demande=workflow.case, actor=workflow.actor
    )
    try:
        app = db.get_app()
    except RuntimeError:
        app = None
    if app:
        db.session.add(notification)

    if user.preferences_notifications in (FLUX_TENDU, None):
        send_notification_by_email(notification)
        notification.sent = True

    return notification


def send_notification_by_email(notification: Notification) -> None:
    user: Profile = notification.user
    subject = "Une action a été réalisée sur une de vos demandes Lab&Co"
    recipients = [user.email]
    ctx = {
        "notification": notification,
        "demande": notification.demande,
        "now": datetime.now(),
    }
    html = render_template("emails/notif-generique.html", **ctx)
    msg = Message(subject, recipients=recipients, html=html)
    try:
        mail.send(msg)
    except SMTPException as e:
        logger.error("SMTP error", e)


def send_email(recipients: set[Profile], subject: str, template, context) -> None:
    assert isinstance(recipients, set)

    recipients = [r.email for r in recipients]
    html = render_template("emails/" + template, **context)
    msg = Message(subject, recipients=recipients, html=html)
    mail.send(msg)



from __future__ import annotations

from typing import Any, Collection

from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method
from marshmallow import Schema, fields

from labster.di import injector
from labster.domain2.model.profile import Profile, ProfileRepository
from labster.domain2.model.structure import Structure, StructureId, \
    StructureRepository
from labster.domain2.model.type_structure import DE, EQ, FA
from labster.domain2.services.roles import Role, RoleService
from labster.rbac import get_permissions_for_structure
from labster.security import get_current_user
from labster.types import JSON, JSONDict
from labster.util import sort_by_name

structure_repo = injector.get(StructureRepository)
profile_repo = injector.get(ProfileRepository)
role_service = injector.get(RoleService)
db = injector.get(SQLAlchemy)


@method
def get_roles(structure_id: str) -> list[dict[str, Any]]:
    structure = structure_repo.get_by_id(StructureId(structure_id))
    assert structure

    role_to_users = role_service.get_users_with_role_on(structure)

    if structure.type in {DE, EQ}:
        roles = [
            Role.RESPONSABLE,
        ]
    else:
        roles = [
            Role.SIGNATAIRE,
            Role.RESPONSABLE,
            Role.ADMIN_LOCAL,
            Role.GESTIONNAIRE,
            Role.PORTEUR,
        ]

    result: list[dict[str, Any]] = []
    for role in roles:
        role_dto = {
            "key": role.name,
            "label": role.value,
            "users": convert_users_to_dto(role_to_users[role]),
        }
        result += [role_dto]
    return result


@method
def get_role_selectors(structure_id: str) -> JSON:
    structure = structure_repo.get_by_id(StructureId(structure_id))
    assert structure

    permissions = get_permissions_for_structure(structure)
    if "P5" not in permissions:
        return []

    is_admin_central = _is_admin_central()
    is_admin_local = _is_admin_local(structure)
    is_admin_facultaire = _is_admin_facultaire(structure)

    if not (is_admin_central or is_admin_local):
        return []

    if structure.type in (DE, EQ):
        roles = [Role.RESPONSABLE]
    elif is_admin_central or is_admin_facultaire:
        roles = [
            Role.SIGNATAIRE,
            Role.RESPONSABLE,
            Role.ADMIN_LOCAL,
            Role.GESTIONNAIRE,
            Role.PORTEUR,
        ]
    else:
        roles = [
            Role.RESPONSABLE,
            Role.ADMIN_LOCAL,
            Role.GESTIONNAIRE,
            Role.PORTEUR,
        ]

    membres = get_membres(structure)
    role_to_users = role_service.get_users_with_role_on(structure)

    result: list[dict[str, Any]] = []
    for role in roles:
        users_with_role = role_to_users[role]
        multiple = role != Role.SIGNATAIRE
        value: JSON
        if multiple:
            value = [{"id": u.id, "label": u.name} for u in users_with_role]
        else:
            if users_with_role:
                u = list(users_with_role)[0]
                value = {"id": u.id, "label": u.name}
            else:
                value = None
        if role != Role.GESTIONNAIRE or not is_admin_central:
            options = [{"id": m.id, "label": m.name} for m in membres]
        else:
            all_users = (
                db.session.query(Profile)
                .filter_by(active=True)
                .order_by(Profile.nom, Profile.prenom)
                .all()
            )
            options = [{"id": m.id, "label": m.name} for m in all_users]

        selector_dto = {
            "key": role.name,
            "label": role.value,
            "value": value,
            "options": options,
            "multiple": multiple,
        }
        result += [selector_dto]

    return result


def _is_admin_central() -> bool:
    current_user = get_current_user()
    if not current_user.is_authenticated:
        return True

    profile = current_user.profile
    return profile.has_role(Role.ADMIN_CENTRAL)


def _is_admin_local(structure: Structure) -> bool:
    current_user = get_current_user()
    if not current_user.is_authenticated:
        return True

    profile = current_user.profile
    for ancestor in [structure] + structure.ancestors:
        if profile.has_role(Role.ADMIN_LOCAL, ancestor):
            return True

    return False


def _is_admin_facultaire(structure: Structure) -> bool:
    current_user = get_current_user()
    if not current_user.is_authenticated:
        return True

    profile = current_user.profile
    for ancestor in [structure] + structure.ancestors:
        if ancestor.type == FA and profile.has_role(Role.ADMIN_LOCAL, ancestor):
            return True

    return False


@method
def get_global_roles() -> JSONDict:
    def make_role_dto(role: Role) -> JSONDict:
        users_with_role = role_service.get_users_with_role(role)
        return {
            "key": role.name,
            "label": role.value,
            "users": convert_users_to_dto(users_with_role),
        }

    roles = [
        make_role_dto(Role.ADMIN_CENTRAL),
        make_role_dto(Role.FAQ_EDITOR),
    ]

    all_users = (
        db.session.query(Profile)
        .filter_by(active=True)
        .order_by(Profile.nom, Profile.prenom)
        .all()
    )
    options = [{"id": m.id, "label": m.name} for m in all_users]

    def make_selector(role: Role):
        users_with_role = role_service.get_users_with_role(role)
        value = [{"id": u.id, "label": u.name} for u in users_with_role]

        return {
            "key": role.name,
            "label": role.value,
            "value": value,
            "options": options,
            "multiple": True,
        }

    selectors = [
        make_selector(Role.ADMIN_CENTRAL),
        make_selector(Role.FAQ_EDITOR),
    ]

    return {
        "roles": roles,
        "selectors": selectors,
    }


#
# Util
#
def convert_users_to_dto(users: Collection[Profile]) -> list[JSON]:
    class ProfileSchema(Schema):
        id = fields.String()
        name = fields.String(attribute="reversed_name")

    users = sort_by_name(users)
    return ProfileSchema().dump(users, many=True).data


def get_membres(structure):
    """Retourne les membres (affectés et rattachés) et de la sous-structure
    (sans les membres affiliés)."""
    role_service = injector.get(RoleService)

    m1 = role_service.get_users_with_given_role(Role.MEMBRE_AFFECTE, structure)
    m2 = role_service.get_users_with_given_role(Role.MEMBRE_RATTACHE, structure)
    membres = list(set(m1) | set(m2))
    membres = sort_by_name(membres)
    membres = [m for m in membres if m.active]
    return membres



from __future__ import annotations

from glom import glom
from pytest import mark

from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure
from labster.domain2.model.type_structure import UN
from labster.domain2.services.roles import Role
from labster.rpc.commands.roles import add_roles, delete_role
from labster.rpc.queries.roles import get_role_selectors, get_roles


@mark.skip
def test_roles(structure_repo, profile_repo):
    universite = Structure(nom="SU", type_name=UN.name)
    structure_repo.put(universite)

    user = Profile(uid="toto")
    profile_repo.put(user)

    result = get_roles(universite.id)
    assert glom(result, (["users"], [["id"]])) == [[], [], [], [], []]

    add_roles(universite.id, [user.id], Role.PORTEUR.name)
    result = get_roles(universite.id)
    assert glom(result, (["users"], [["id"]])) == [[], [], [], [], [user.id]]

    delete_role(universite.id, user.id, Role.PORTEUR.name)
    result = get_roles(universite.id)
    assert glom(result, (["users"], [["id"]])) == [[], [], [], [], []]

    result = get_role_selectors(universite.id)
    # First select is not multiple
    assert glom(result, ["value"]) == [None, [], [], [], []]



from __future__ import annotations

import threading
from typing import Any

from jsonrpcserver import method
from marshmallow import Schema, fields

from labster.di import injector
from labster.domain2.model.notification import Notification
from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import StructureRepository
from labster.domain2.services.constants import get_constant, get_constants
from labster.domain2.services.demande import get_demande_types_for_user
from labster.domain2.services.roles import Role
from labster.menu import get_menu
from labster.rbac import is_membre_dri, is_membre_drv
from labster.security import get_current_profile
from labster.types import JSONDict

from .demandes_tables import mes_taches, mes_taches_en_retard
from .home_boxes import get_boxes

structure_repo = injector.get(StructureRepository)

cache = threading.local()


@method
def get_user_context() -> JSONDict:
    user = get_current_profile()
    return get_context(user)


def get_context(user: Profile) -> JSONDict:
    menus = get_menu(user)
    types_demandes = list(get_demande_types_for_user(user))

    is_responsable = user.has_role(Role.RESPONSABLE, "*")
    is_porteur = user.has_role(Role.PORTEUR, "*")
    is_gestionnaire = user.has_role(Role.GESTIONNAIRE, "*")

    message_dgrtt = get_constant("message_dgrtt", "OK")

    return {
        "menu": [menu.asdict() for menu in menus],
        "user": UserSchema().dump(user).data,
        "types_demandes": types_demandes,
        "home_boxes": get_boxes(),
        "archives_boxes": get_boxes(archives=True),
        "is_admin": user.has_role(Role.ADMIN_CENTRAL),
        "is_membre_dri": is_membre_dri(user),
        "is_membre_drv": is_membre_drv(user),
        "is_responsable": is_responsable,
        "is_porteur": is_porteur,
        "is_gestionnaire": is_gestionnaire,
        "message_dgrtt": message_dgrtt,
        "constants": get_constants(),
    }


class UserSchema(Schema):
    nom = fields.String()
    prenom = fields.String()
    uid = fields.String()
    id = fields.String()
    login = fields.String()

    is_admin = fields.Method("get_is_admin")
    roles = fields.Method("get_roles")

    nb_notifications_non_vues = fields.Method("get_nb_notifications_non_vues")
    nb_taches = fields.Method("get_nb_taches")
    nb_taches_retard = fields.Method("get_nb_taches_retard")

    def get_is_admin(self, user: Profile):
        return user.has_role(Role.ADMIN_CENTRAL)

    def get_roles(self, user: Profile) -> list[Any]:
        assert isinstance(user, Profile)
        return []

    def get_nb_notifications_non_vues(self, user: Profile) -> int:
        date = user.date_derniere_notification_vue

        query = Notification.query.filter(Notification.user == user)
        if date:
            query = query.filter(Notification.created_at > date)

        return query.count()

    def get_nb_taches_retard(self, user: Profile) -> int:
        return len(mes_taches_en_retard(user))

    def get_nb_taches(self, user: Profile) -> int:
        return len(mes_taches(user))



"""Create an application instance."""
from pathlib import Path


from labster.app import create_app

app = create_app()


newrelic_config = Path(__file__).parent / "newrelic.ini"
if newrelic_config.exists():
    import newrelic.agent

    newrelic.agent.initialize(newrelic_config, environment="production")

    app = newrelic.agent.WSGIApplicationWrapper(app)



# Import old stuff for now
from __future__ import annotations

from . import config  # noqa
from .financeurs import *  # noqa
from .mapping_dgrtt import *  # noqa
from .notifications import *  # noqa
from .roles import *  # noqa



from __future__ import annotations

import json
from datetime import datetime
from urllib.parse import parse_qs, urlencode, urljoin, urlsplit, urlunsplit

import requests
import structlog
from flask import Flask, Request, abort, redirect, render_template, request, \
    session, url_for
from flask_sqlalchemy import SQLAlchemy
from werkzeug.exceptions import Unauthorized

from labster.di import injector
from labster.domain2.model.profile import Profile
from labster.domain2.services.constants import get_constants
from labster.extensions import login_manager
from labster.security import get_current_user

from . import route

logger = structlog.get_logger()

db = injector.get(SQLAlchemy)


@login_manager.unauthorized_handler
def unauthorized(app: Flask, request: Request):
    accept_header = request.headers.get("Accept", "")
    if "application/json" in accept_header or app.testing:
        abort(401)

    return cas_login(app.config["CAS_SERVER"])


#
# Routes
#
@route("/login")
def login():
    current_user = get_current_user()

    if not current_user.is_authenticated:
        if _single_user():
            return render_template("auth/single_user.j2")
        return render_template("auth/login_cas.j2")

    return render_template("auth/redirect.j2")


def _single_user():
    if "bypass" in request.args:
        return False

    constants = get_constants()
    value = str(constants.get("single_user")).lower()
    return value in {"none", "true", "y"}


@route("/login", methods=["POST"])
def login_post():
    if "current_user_id" in session:
        del session["current_user_id"]
    return redirect(url_for(".login", _external=True))


@route("/go")
def go(app: Flask, request: Request):
    accept_header = request.headers.get("Accept", "")
    if "application/json" in accept_header or app.testing:
        abort(401)

    return cas_login(app.config["CAS_SERVER"])


@route("/callback")
def callback(app: Flask, request: Request):
    ticket = request.args.get("ticket", "")
    default_url = url_for(".login", _external=True)
    next_url = request.args.get("next", default_url)
    if not login_with_ticket(ticket, app.config["CAS_SERVER"]):
        login_manager.unauthorized()

    return redirect(next_url)


#
# Util
#
def cas_login(cas_server: str):
    url = urljoin(cas_server, "login")
    scheme, netloc, path, query, fragment = urlsplit(url)
    args = parse_qs(query)
    args["service"] = [url_for("auth.callback", _external=True)]
    query = urlencode(args, doseq=True)
    url = urlunsplit((scheme, netloc, path, query, fragment))
    return redirect(url)


def login_with_ticket(ticket: str, cas_server: str):
    url = urljoin(cas_server, "p3/serviceValidate")
    service_url = url_for("auth.callback", _external=True)
    params = {"service": service_url, "ticket": ticket, "format": "JSON"}
    session = requests.Session()
    r = session.get(url, params=params)

    r.raise_for_status()

    if r.status_code != 200:
        logger.debug(
            "Error during CAS ticket validation:\nresponse code: %d"
            '\nContent: """\n%s\n"""',
            r.status_code,
            r.content,
        )
        raise ValueError(
            f"Error during CAS ticket validation reponse code: {r.status_code}"
        )
    service_response = r.json()["serviceResponse"]

    if "authenticationSuccess" not in service_response:
        # TODO: get attr "code" for details
        logger.debug(r.content)
        raise Unauthorized()

    auth_data = service_response["authenticationSuccess"]

    # Note: this is the "old" uid
    login = auth_data["user"]
    attributes = auth_data["attributes"]

    if login == "fermigier":
        login = "poulainm"

    # FIXME: uid is not the "new" uid (which is attributes["uid"])
    user = get_user_by_login(login)

    if not user:
        raise Unauthorized()
        # user = Profile(uid=uid)
        # user.nom = attributes["sn"]
        # user.prenom = attributes["givenName"]
        # # user.roles = attributes["eduPersonAffiliation"]
        #
        # db.session.add(user)

    login_user(user)

    # TODO: add these fields to Profile
    user.cas_entry = json.dumps(attributes)
    user.date_last_login = datetime.utcnow()

    db.session.commit()
    return redirect(url_for(".login", _external=True))


def get_user_by_login(login: str) -> Profile:
    # This fixes some nasty "current transaction is aborted" bug
    db.session.commit()
    query = db.session.query(Profile).filter_by(active=True)
    user = query.filter_by(login=login).first()
    return user


def login_user(user):
    assert user.active
    session["current_user_id"] = user.id
    logger.info("User has logged in", login=user.login, uid=user.uid)



"""Integration tests."""
from __future__ import annotations

import os
import uuid

import pytest
from flask import url_for

from labster.di import injector
from labster.domain2.model.structure import Structure, StructureRepository
from labster.domain2.model.type_structure import UN
from labster.domain.models.profiles import Profile
from labster.domain.models.unites import LABORATOIRE, OrgUnit
from labster.domain.services.roles import ROLES

IGNORED_ENDPOINTS = ["auth.callback", "auth.backdoor"]


# FIXME
if os.environ.get("TRAVIS"):
    pytest.skip("These tests currently fail on Travis", allow_module_level=True)

pytest.skip(
    "These tests currently fail since recent refactoring", allow_module_level=True
)

structure_repo = injector.get(StructureRepository)


#
# 'auth' blueprint
#
def test_login(client, db_session):
    r = client.get(url_for("auth.login"))
    assert r.status_code == 200


#
# 'main' blueprint
#
def test_home_as_anonymous(client, db_session):
    r = client.get(url_for("main.home"))
    assert r.status_code == 302


def test_home_as_dgrtt(client, db_session):
    login_as_dgrtt(client, db_session)
    r = client.get(url_for("main.home"), follow_redirects=True)
    assert r.status_code == 200


def test_home_as_gdl(client, db_session):
    login_as_gdl(client, db_session)
    r = client.get(url_for("main.home"), follow_redirects=True)
    assert r.status_code == 200


#
#
#
def test_url_for(app, db_session):
    assert url_for("main.home") == "http://localhost.localdomain/"

    p = Profile(uid="ayuso", nom="Ayuso", prenom="Sandrine", email="ayuso@example.com")
    assert url_for(p)

    dn = f"test{uuid.uuid4()}"
    nom = f"nom{uuid.uuid4()}"
    labo = OrgUnit(type=LABORATOIRE, dn=dn, nom=nom)
    assert url_for(labo)


# TODO: remove soon
# def test_directory(client, db_session):
#     login_as_dgrtt(client, db_session)
#     r = client.get(url_for("directory.home"))
#     assert r.status_code == 200
#
#     login_as_gdl(client, db_session)
#     r = client.get(url_for("main.home"), follow_redirects=True)
#     assert r.status_code == 200
#
#
# def xxx_test_nouvelle_demande_rh(client, db_session):
#     login_as_gdl(client, db_session)
#
#     r = client.get(url_for("demandes.demande_new"))
#     assert r.status_code == 200
#
#     data = {"prenom": "Snake", "nom": "Plisken"}
#     r = client.post(
#         url_for("demandes.demande_new_post"), data=data, follow_redirects=True
#     )
#     assert r.status_code == 200
#     assert "Snake Plisken" in r.get_data(as_text=True)
#
#     # debug_info = get_debug_info(r)
#     # url = debug_info['url']
#
#     r = client.get(url_for("demandes.demandes"), follow_redirects=True)
#     assert r.status_code == 200
#     assert "Snake Plisken" in r.get_data(as_text=True)
#
#
# #
# # Admin
# #
# def test_admin_views(client, db_session):
#     login_as_dgrtt(client, db_session)
#
#     r = client.get(url_for("admin2.home"))
#     assert r.status_code == 200
#
#     r = client.get(url_for("admin2.financeurs"))
#     assert r.status_code == 200
#
#     r = client.get(url_for("admin2.mapping_dgrtt"))
#     assert r.status_code == 200


#
# All...
#
@pytest.mark.parametrize("role", ROLES)
def test_all_simple_endpoints(role, client, app, db_session):
    # Setup repo with a root
    structure_repo.clear()
    universite = Structure(
        nom="Sorbonne Université",
        type_name=UN.name,
        sigle="SU",
        dn="ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    )
    structure_repo.put(universite)

    login_as_dgrtt(client, db_session)

    Profile.__roles = [role]

    endpoints = get_endpoints(app)
    try:
        for endpoint in endpoints:
            print(f"checking endpoint '{endpoint}' with role '{role}'")
            check_endpoint(endpoint, client)
    finally:
        del Profile.__roles

    structure_repo.clear()


def get_endpoints(app):
    endpoints = []
    for rule in app.url_map.iter_rules():
        if "GET" not in rule.methods:
            continue

        endpoint = rule.endpoint
        blueprint_name = endpoint.split(".")[0]

        # Skip a few hard cases (FIXME after the SPA migration)
        if blueprint_name not in {"auth", "v3"}:
            continue

        # if blueprint_name not in ["main", "admin2"]:
        #     return
        if endpoint in IGNORED_ENDPOINTS:
            continue
        if rule.arguments:
            continue

        endpoints.append(rule.endpoint)

    return endpoints


def check_endpoint(endpoint, client):
    url = url_for(endpoint)
    try:
        r = client.get(url)
    except Exception:
        print(f"Error on url: {url} (endpoint: {endpoint})")
        raise
    assert r.status_code in (200, 302, 401, 403), f"for endpoint = '{endpoint}'"


#
# Util
#
def login_as_dgrtt(client, db_session):
    p = Profile(uid="ayuso", nom="Ayuso", prenom="Sandrine", email="ayuso@example.com")
    p.__roles = ["alc", "dgrtt"]
    db_session.add(p)
    db_session.flush()

    login_as(p, client)


def login_as_gdl(client, db_session):
    dn = f"test{uuid.uuid4()}"
    nom = f"nom{uuid.uuid4()}"
    labo = OrgUnit(type=LABORATOIRE, dn=dn, nom=nom)
    p = Profile(
        uid="courtoisi",
        nom="Courtois",
        prenom="Isabelle",
        email="courtoisi@example.com",
        laboratoire=labo,
    )
    db_session.add(p)
    db_session.flush()

    id = p.id
    p1 = Profile.query.get(id)
    assert p1 is p

    login_as(p, client)


def login_as(profile: Profile, client):
    uid = profile.uid
    r = client.get(url_for("auth.backdoor", uid=uid))
    assert r.status_code == 201

    r = client.get(url_for("v3.self"))
    assert r.status_code == 200
    assert r.json["data"]["uid"] == uid



from __future__ import annotations

from marshmallow import fields
from marshmallow_annotations import registry

from labster.domain2.model.profile import ProfileId
from labster.domain2.model.structure import StructureId

registry.register_field_for_type(StructureId, fields.Integer)
registry.register_field_for_type(ProfileId, fields.Integer)



# from __future__ import annotations
#
#
# class BaseInMemoryRepository:
#     def __init__(self):
#         self._data = {}
#
#     def clear(self):
#         self._data.clear()
#
#     def is_empty(self):
#         return len(self._data) == 0
from __future__ import annotations



"""
Auth blueprint - used for authencation.
"""
from __future__ import annotations

from flask import Blueprint

blueprint = Blueprint("auth", __name__, template_folder="templates", url_prefix="")
route = blueprint.route

__all__ = ()


@blueprint.record
def configure(state):
    from . import backdoors, cas



from __future__ import annotations

from datetime import datetime

import pytz
from abilian.core.models import IdMixin
from flask_sqlalchemy import BaseQuery
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, \
    String, Unicode
from sqlalchemy.orm import relationship

from labster.extensions import db

from .demande import Demande
from .profile import Profile


class NotificationQuery(BaseQuery):
    def get_for_user(self, user, page=0):
        # TODO: paging
        # FIXME
        return (
            self.filter(Notification.user == user)
            .order_by(Notification.created_at.desc())
            .limit(25)
            .all()
        )


class Notification(IdMixin, db.Model):
    __tablename__ = "v3_notifications"
    query_class = NotificationQuery

    created_at = Column(DateTime, default=datetime.utcnow)

    body = Column(Unicode, nullable=False)

    user_id = Column(String(36), ForeignKey(Profile.id))
    user = relationship(Profile, foreign_keys=[user_id])

    actor_id = Column(String(36), ForeignKey(Profile.id))
    actor = relationship(Profile, foreign_keys=[actor_id])

    demande_id = Column(Integer, ForeignKey(Demande.id))
    demande = relationship(Demande, foreign_keys=[demande_id])

    sent = Column(Boolean, default=False)

    @property
    def created_at_tz(self):
        paris = pytz.timezone("Europe/Paris")
        return paris.localize(self.created_at)



from __future__ import annotations

from flask import flash, request

from labster.extensions import db
from labster.security import get_current_profile

from .. import route

CHOICES = [
    (0, "En flux continu (défaut)"),
    (1, "Agrégées, une fois par jour"),
    (2, "Agrégées, une fois par semaine (le jeudi)"),
]


@route("/user/preferences")
def preferences():
    user = get_current_profile()

    if user.preferences_notifications is None:
        user.preferences_notifications = 0

    if user.preferences_nb_jours_notifications is None:
        user.preferences_nb_jours_notifications = 0

    return {
        "choices": CHOICES,
        "preferences_notifications": user.preferences_notifications,
        "nb_jours_notification": user.preferences_nb_jours_notifications,
    }


@route("/user/preferences", methods=["POST"])
def preferences_post():
    user = get_current_profile()

    payload = request.json

    pref = payload["preferences_notifications"]
    nb_jours_notification = payload["nb_jours_notification"]

    modified = False
    if pref != user.preferences_notifications:
        user.preferences_notifications = pref
        modified = True

    if nb_jours_notification != user.preferences_nb_jours_notifications:
        user.preferences_nb_jours_notifications = nb_jours_notification
        modified = True

    if modified:
        db.session.commit()
        flash("Vos préférences ont été mises à jours")

    return "", 204



"""Tentative de modéliser les rôles de l'application sous forme de hiérarchie
de classes."""

from __future__ import annotations

import enum

from abilian.core.entities import Entity
from attr import attrib, attrs
from sqlalchemy import Column, Enum, ForeignKey, Integer
from sqlalchemy.orm import backref, relationship

from .profiles import Profile
from .unites import OrgUnit


class RoleType(enum.Enum):
    MEMBRE = "Membre"
    GDL = "Gestionnaire de demande Labster"
    DIRECTION = "Direction"
    PORTEUR = "Porteur"
    ALL = "Administrateur Labster local"

    CONTACT_DGRTT = "Contact DGRTT"
    CHEF_DE_BUREAU_DGRTT = "Chef de bureau DGRTT"
    ALC = "Administrateur Labster central"


TYPE_ENUM = [t.value for t in list(RoleType)]


@attrs(init=False, these={"type": attrib(), "context": attrib()}, hash=True)
class Role(Entity):
    # __tablename__ = 'role'
    __indexable__ = False

    type = Column(Enum(*TYPE_ENUM, name="role_type"), nullable=False)

    profile_id = Column(Integer, ForeignKey(Profile.id), index=True)
    profile = relationship(
        Profile,
        foreign_keys=[profile_id],
        backref=backref("roles", cascade="all,delete"),
    )

    context_id = Column(Integer, ForeignKey(OrgUnit.id), index=True)
    context = relationship(
        OrgUnit,
        foreign_keys=[context_id],
        backref=backref("contexts", cascade="all,delete"),
    )

    # def can(self, action, target):
    #   pass

    def __unicode__(self):
        if self.context:
            return "<{} de la structure: {} ({})>".format(
                self.type, self.context.nom, self.context.type
            )
        else:
            return f"<{self.type}>"



from __future__ import annotations

from datetime import date
from typing import Any

import structlog

logger = structlog.get_logger()

SEP = ("-", "/")


def parse_date(dd: Any) -> date | None:
    if not dd:
        return None

    try:
        if dd[4] in SEP and dd[7] in SEP:
            year = int(dd[0:4])
            month = int(dd[5:7])
            day = int(dd[8:10])
            value = date(year, month, day)
            return value
        elif dd[2] in SEP and dd[5] in SEP:
            day = int(dd[0:2])
            month = int(dd[3:5])
            year = int(dd[6:10])
            value = date(year, month, day)
            return value
        else:
            return None
    except Exception:
        logger.exception(f"Can't parse date: {dd}")
        return None



from __future__ import annotations

import flask_mail
import html2text
from flask import current_app
from flask_mail import Message

# logger = logging.getLogger(__name__)


class Mail(flask_mail.Mail):
    def send(self, message: Message) -> None:
        config = current_app.config

        substitute_address = config.get("TEST_EMAIL_ADRESS")
        if substitute_address:
            assert isinstance(substitute_address, str)
            message.subject += f" [pour: {', '.join(message.recipients)}]"
            message.recipients = [substitute_address]
            message.cc = []
            message.bcc = []

        if not message.body:
            message.body = html2text.html2text(message.html)

        if not message.sender:
            message.sender = config["MAIL_SENDER"]

        email_cc = config.get("EMAIL_CC")
        if email_cc:
            if isinstance(email_cc, list) or isinstance(email_cc, tuple):
                message.bcc = email_cc
            else:
                assert isinstance(email_cc, str)
                message.bcc = [email_cc]

        # if message.sender and config[""]:
        #     flask_mail.Mail.send(self, message)
        flask_mail.Mail.send(self, message)



from __future__ import annotations

from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method

from labster.di import injector
from labster.domain2.model.profile import ProfileRepository
from labster.domain2.model.structure import StructureRepository
from labster.domain2.services.contacts import ContactService, ContactType
from labster.rbac import check_can_edit_contacts

structure_repo = injector.get(StructureRepository)
profile_repo = injector.get(ProfileRepository)
contact_service = injector.get(ContactService)
db = injector.get(SQLAlchemy)


@method
def update_contacts(structure_id: str, contacts: dict[str, str]):
    structure = structure_repo.get_by_id(structure_id)
    check_can_edit_contacts(structure)

    for contact_type_name, uid in contacts.items():
        contact_type = getattr(ContactType, contact_type_name)
        if uid:
            profile = profile_repo.get_by_uid(uid)
            contact_service.set_contact(structure, contact_type, profile)
        else:
            contact_service.delete_contact(structure, contact_type)

    db.session.commit()



"""Configuration and injectable fixtures for Pytest."""
from __future__ import annotations

import logging
import os

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from pytest import fixture
from sqlalchemy.exc import SQLAlchemyError
from typeguard import TypeChecker

from labster.app import create_app
from labster.domain.services.constants import get_initial_constants
from labster.extensions import db as _db

checker = TypeChecker("labster")

if "TYPECHECK" in os.environ:
    logging.captureWarnings(True)
    if not checker.active:
        checker.start()


class TestConfig:
    TESTING = True
    CSRF_ENABLED = False
    MAIL_SENDER = "test@example.com"
    MAIL_SUPPRESS_SEND = True
    SECRET_KEY = "changeme"
    SERVER_NAME = "localhost.localdomain"
    SQLALCHEMY_DATABASE_URI = "sqlite://"


@fixture(scope="session")
def app() -> Flask:
    """We usually only create an app once per session."""

    return create_app(TestConfig)


@fixture
def app_context(app: Flask):
    with app.app_context() as ctx:
        yield ctx


@fixture
def request_context(app: Flask):
    with app.test_request_context() as ctx:
        yield ctx


@fixture
def db(app: Flask) -> SQLAlchemy:
    """Return a fresh db for each test."""

    with app.app_context():
        cleanup_db(_db)
        _db.create_all()

        yield _db

        _db.session.remove()
        cleanup_db(_db)

        _db.session.flush()


@fixture
def db_session(db: SQLAlchemy):
    """Kept for historical reasons."""

    return db.session


@fixture
def config():
    from labster.domain.models.config import Config

    DATA = get_initial_constants()
    config = Config()
    config.data = DATA
    return config


@fixture
def client(app, db):
    """Return a Web client, used for testing, bound to a DB session."""
    return app.test_client()


@fixture
def injector(app):
    from labster.di import injector

    return injector


#
# Cleanup utilities
#
def cleanup_db(db):
    """Drop all the tables, in a way that doesn't raise integrity errors."""

    for table in reversed(db.metadata.sorted_tables):
        try:
            db.session.execute(table.delete())
        except SQLAlchemyError:
            print(f"Failed to delete table {table}")
            pass



from __future__ import annotations

from functools import singledispatch

import requests
from flask import current_app
from marshmallow import Schema, fields

from labster.domain2.model.demande import Demande
from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure
from labster.extensions import db
from labster.signals import model_saved

COLLECTIONS = ["users", "structures", "demandes"]


class ProfileSchema(Schema):
    uid = fields.Str()
    full_name = fields.Str()
    fonction_structurelle_principale = fields.Str()
    ldap_dict = fields.Dict()


class PorteurSchema(ProfileSchema):
    pass


class GestionnaireSchema(ProfileSchema):
    pass


class OrgUnitSchema(Schema):
    nom = fields.Str()
    sigle = fields.Str()
    dn = fields.Str()


class DemandeSchema(Schema):
    id = fields.Integer()
    nom = fields.Str()
    no_infolab = fields.Str()
    no_eotp = fields.Str()

    # FIXME
    # date_effective = fields.DateTime()
    # date_soumission = fields.DateTime()
    # date_finalisation = fields.DateTime()

    porteur = fields.Nested(PorteurSchema)
    gestionnaire = fields.Nested(GestionnaireSchema)
    contact_dgrtt = fields.Nested(ProfileSchema)

    laboratoire = fields.Nested(OrgUnitSchema)
    structure = fields.Nested(OrgUnitSchema)

    active = fields.Boolean()
    wf_state = fields.Str()
    wf_date_derniere_action = fields.DateTime()
    wf_retard = fields.Integer()
    wf_current_owner = fields.Str()
    wf_history = fields.List(fields.Dict())
    age = fields.Integer()
    retard = fields.Integer()
    contact = fields.Str()

    data = fields.Method("get_data")

    def get_data(self, obj):
        d = obj.data
        d = {k: v for k, v in d.items() if not k.startswith("html-")}
        return d


class RestHeart:
    def __init__(self):
        config = current_app.config
        self.root_url = config.get("RESTHEART_URL")
        self.auth = config["RESTHEART_AUTH"]

    def put(self, url, **kw):
        url = self.root_url + "/" + url
        response = requests.put(url, auth=self.auth, **kw)
        return response

    def post(self, url, **kw):
        url = self.root_url + "/" + url
        response = requests.post(url, auth=self.auth, **kw)
        return response


def init():
    rh = RestHeart()
    rh.put("db")
    for coll_name in COLLECTIONS:
        print(f"Creating collection {coll_name}")
        response = rh.put(coll_name)
        if response.status_code not in (200, 201):
            raise RuntimeError(f"Server send invalid response: {response}")


def sync_all_objects():
    print("Syncing users")
    sync_objects(Profile)
    print("Syncing structures")
    sync_objects(Structure)
    print("Syncing demandes")
    sync_objects(Demande)


def sync_objects(cls: type):
    objects = db.session.query(cls).all()
    for obj in objects:
        sync_object(obj)


def sync_object(obj):
    rh = RestHeart()
    url, data = serialize(obj)
    # print(url, data)
    response = rh.put(url, json=data)
    if response.status_code not in (200, 201):
        raise RuntimeError(f"Server send invalid response: {response}")


@singledispatch
def serialize(obj):
    raise NotImplementedError


@serialize.register(Profile)
def serialize_profile(obj):
    url = f"users/{obj.uid}"
    schema = ProfileSchema()
    result = schema.dump(obj)
    assert not result.errors, result.errors
    data = result.data
    return url, data


@serialize.register(Structure)
def serialize_structure(obj):
    url = f"structures/{obj.id}"
    schema = OrgUnitSchema()
    result = schema.dump(obj)
    assert not result.errors, result.errors
    data = result.data
    return url, data


@serialize.register(Demande)
def serialize_demande(obj):
    url = f"demandes/{obj.id}"
    schema = DemandeSchema()
    result = schema.dump(obj)
    assert not result.errors, result.errors
    data = result.data
    return url, data


def on_model_saved(obj):
    print(f"Model {obj} saved")
    sync_object(obj)


def register_callback() -> None:
    model_saved.connect(on_model_saved)



from __future__ import annotations

import json
from typing import Any

from werkzeug.exceptions import NotFound

from labster.di import injector
from labster.domain2.model.profile import Profile, ProfileId, ProfileRepository
from labster.domain2.model.structure import Structure, StructureRepository
from labster.domain2.model.type_structure import DE, EQ
from labster.domain2.services.contacts import ContactService
from labster.domain2.services.roles import Role, RoleService
from labster.rpc.registry import context_for
from labster.types import JSON, JSONDict

role_service = injector.get(RoleService)
contact_service = injector.get(ContactService)
profile_repo = injector.get(ProfileRepository)
structure_repo = injector.get(StructureRepository)


@context_for("user")
def get_user(id: str) -> JSONDict:
    user = profile_repo.get_by_id(ProfileId(id))
    if not user:
        user = profile_repo.get_by_old_uid(id)
    if not user:
        raise NotFound()

    structure_affectation_dto: JSON
    if user.affectation:
        structure_affectation = structure_repo.get_by_dn(user.affectation)
        if structure_affectation:
            structure_affectation_dto = {
                "name": structure_affectation.sigle_ou_nom,
                "id": structure_affectation.id,
            }
        else:
            structure_affectation_dto = None
    else:
        structure_affectation_dto = None

    user_dto = {}
    for k in ["nom", "prenom", "email", "telephone", "uid", "affectation", "login"]:
        user_dto[k] = getattr(user, k)

    # Cas particulier: fonctions est un set() pas une liste.
    if isinstance(user.fonctions, str):
        fonctions = json.loads(user.fonctions)
    else:
        fonctions = user.fonctions
    user_dto["fonctions"] = sorted(fonctions)

    roles_dto = get_roles_dto_for_user(user, skip=True)
    perimetre_dto = get_perimetre_dto_for_user(user)

    ctx = {
        "name": user.full_name,
        "user": user_dto,
        "structure_affectation": structure_affectation_dto,
        "roles": roles_dto,
        "perimetre": perimetre_dto,
    }
    return ctx


#
# Serialization helpers
#
def get_roles_dto_for_user(
    user: Profile, base_structure: Structure | None = None, skip: bool = False
) -> list[dict[str, Any]]:
    roles_for_user = role_service.get_roles_for_user(user)

    all_structures = {}
    for contexts in roles_for_user.values():
        for context in contexts:
            if not context:
                continue
            assert isinstance(context, Structure)
            all_structures[context.id] = context

    list_structures = list(all_structures.values())
    list_structures.sort(key=lambda s: s.depth)

    ancestors: set[Structure]
    if base_structure:
        ancestors = set(base_structure.ancestors)
    else:
        ancestors = set()

    def get_roles_list(structure):
        role_list: list[str] = []
        for role, structures in roles_for_user.items():
            if role == Role.MEMBRE:
                continue

            if structure in ancestors:
                continue

            if (
                skip
                and role == Role.MEMBRE_AFFILIE
                and structure.type_name != "Département"
            ):
                continue

            if structure in structures:
                role_list.append(role.value)

        role_set = set(role_list)
        if Role.MEMBRE_AFFECTE.value in role_set:
            role_set.discard(Role.MEMBRE_AFFILIE.value)
        role_list = list(role_set)

        role_list.sort()
        return role_list

    roles_dto = []
    for structure in list_structures:
        role_list = get_roles_list(structure)
        if not role_list:
            continue

        dto = {
            "structure": {
                "name": structure.sigle_ou_nom,
                "type": structure.type_name,
                "reelle": structure.is_reelle,
                "id": structure.id,
                "depth": structure.depth,
            },
            "roles": role_list,
        }
        roles_dto.append(dto)

    def sorter(dto) -> tuple[int, int]:
        structure = dto["structure"]
        depth = structure["depth"]
        if set(dto["roles"]) & {
            Role.MEMBRE_AFFECTE.value,
            Role.MEMBRE_RATTACHE.value,
            Role.MEMBRE_AFFILIE.value,
            Role.MEMBRE.value,
            Role.PORTEUR.value,
            Role.SIGNATAIRE.value,
            Role.RESPONSABLE.value,
        }:
            if structure["reelle"]:
                return (1, depth)
            if structure["type"] in {EQ.name, DE.name}:
                return (2, depth)
            return (3, depth)
        if dto["roles"] == Role.GESTIONNAIRE.value:
            return (4, depth)
        if dto["roles"] == Role.ADMIN_LOCAL.value:
            return (5, depth)
        return (6, depth)

    roles_dto.sort(key=sorter)
    return roles_dto


def get_perimetre_dto_for_user(user) -> list[dict[str, Any]]:
    mapping = contact_service.get_mapping()

    result = []
    for structure, d in mapping.items():
        for contact_type, profile in d.items():
            if profile == user:
                structure_dto = {
                    "name": structure.sigle_ou_nom,
                    "id": structure.id,
                    "type": structure.type_name,
                    "depth": structure.depth,
                }
                dto = {"structure": structure_dto, "types": [contact_type.value]}
                result.append(dto)

    return result
